import json
from copy import deepcopy
from pathlib import Path
from typing import List

from otlmow_model.OtlmowModel.BaseClasses.OTLObject import OTLObject
from otlmow_visuals.PyVisWrapper import PyVisWrapper

from otlmow_gui.Domain.ProgramFileStructure import ProgramFileStructure
from otlmow_gui.Domain.logger.OTLLogger import OTLLogger



class VisualisationHelper:
    object_count_limit = 100000

    @classmethod
    def get_std_vis_wrap_instance(cls):
        return PyVisWrapper()

    @classmethod
    def create_html(cls,html_loc:Path,
                    objects_in_memory: List[OTLObject],
                    vis_mode="1 Hiërarchische visualisatie",
                    collection_threshold=-1):


        objects_in_memory = deepcopy(objects_in_memory)
        # sort on typeURI to ensure color coding is always the same for the same set of assets
        objects_in_memory.sort(key=lambda otl_object: otl_object.typeURI)

        visualisation_option = 1
        if vis_mode == "1 Hiërarchische visualisatie":
            visualisation_option = 1
        elif vis_mode == "2 Spinnenweb visualisatie":
            visualisation_option= 4
        elif vis_mode == "3 Shell visualisatie":
            visualisation_option = 3
        elif vis_mode == "4 Repulsion visualisatie":
            visualisation_option = 4
        elif vis_mode == "4 ForceAtlas2Based visualisatie":
            visualisation_option = 5



        stdVis = cls.get_std_vis_wrap_instance()
        stdVis.show(list_of_objects=objects_in_memory, visualisation_option = visualisation_option,
                    html_path=Path(html_loc), launch_html=False,collection_threshold=collection_threshold)

        cls.modify_html(html_loc)

        return stdVis

    @classmethod
    def modify_html(cls, file_path: Path) -> None:
        with open(file_path) as file:
            file_data = file.readlines()

        replace_index_lib = next((
                index for index, line in enumerate(file_data)
                if '<script src="lib/bindings/utils.js"></script>' in line),
            -1)
        if replace_index_lib > 0:
            cls.replace_and_add_lines(file_data, replace_index_lib,
                                      '<script src="lib/bindings/utils.js"></script>',
                                      '<script src="lib/bindings/utils.js"></script>',

                                      ['<script src="qrc:///qtwebchannel/qwebchannel.js"></script>'])

        replace_index = next((
                index for index, line in enumerate(file_data)
                if "drawGraph();" in line),
            -1)
        if replace_index > 0:
            path_for_in_js = str(file_path.absolute()).replace("\\","\\\\")

            add_data = ["var container = document.getElementById('mynetwork');",
                        f'var html_path = "{path_for_in_js}";',
                        "",
                        "// add webchannel to javascript to communicate with python",
                        'document.addEventListener("DOMContentLoaded", function() '
                        '{',

                        "   network.on('stabilizationIterationsDone',onIterationStabilized)",
                        "   network.on('stabilized',onStabilized)",
                        "   container.addEventListener('mouseover', disablePhysics);",
                        "   setTimeout(disablePhysics, 300);",
                        "   makeMeMultiSelect(container, network, nodes);",
                        "   try",
                        "   {",
                        """     new QWebChannel(qt.webChannelTransport, function(channel) 
                                {
                                    window.backend = channel.objects.backend;
                                    
                                    if (window.backend)   
                                    {window.backend.receive_network_loaded_notification();}
                                    else
                                    { 
                                       alert('window.backend.receive_network_loaded_notification() doesnt exist yet');
                                       console.log('QWebChannel is not initialized yet.');
                                    }
                                    
                                });""",
                        "   } ",
                        "   catch (error) ",
                        "   {",
                        "       console.error(error);",
                        # '       alert("DataVisualisationScreen:Error in webchannel creation: " + error);',
                        "   }",

                        "})"]

            add_data.extend(cls.create_disablePhysics_js_function_and_add_to_event())
            add_data.extend(cls.create_ExtractNodeList_js_function())
            add_data.extend(cls.create_ExtractEdgeList_js_function())
            add_data.extend(cls.create_sendCurrentCombinedDataToPython_js_function())
            add_data.extend(cls.create_AddEdge_js_function())
            add_data.extend(cls.create_AddEdgeWithLabel_js_function())
            add_data.extend(cls.create_removeEdge_js_function())
            add_data.extend(cls.create_removeEdgeJointNode_js_function())
            add_data.extend(cls.create_UpdateCollectionAttributes_js_function())
            add_data.extend(cls.create_getFlatListOfRelationIdsInCollection())
            add_data.extend(cls.create_UpdateAllRelationHiddenStatesOfColor_js_function())
            add_data.extend(cls.create_setAllNonCollectionEdgesToUnhidden_js_function())


            add_data.extend(cls.load_js_script_file("dragMultiSelect.js"))

            cls.replace_and_add_lines(file_data, replace_index,
              "drawGraph();",
              "var network = drawGraph();\n"
                "var loadingBar = document.getElementById('loadingBar');\n"
                "if (loadingBar) { loadingBar.style.display = 'none'; }"
                ,add_data)

        with open(file_path, 'w') as file:
            for line in file_data:
                file.write(line)

    @classmethod
    def load_js_script_file(cls, script_filename):
        library_path = ProgramFileStructure.get_dynamic_library_path("javascripts_visualisation")
        script_path = library_path / script_filename
        with open(script_path) as js_script_file:
            js_script_lines = js_script_file.readlines()
            return js_script_lines
        return []

    @classmethod
    def replace_and_add_lines(cls, file_data, replace_index, start_line_to_replace: str,
                              start_replacement: str, list_of_followup_lines: list[str]):
        file_data[replace_index] = file_data[replace_index].replace(start_line_to_replace,
                                                                    start_replacement)
        for i, followup_line in enumerate(list_of_followup_lines):
            file_data.insert(replace_index + i, followup_line + "\n")

    @classmethod
    def create_disablePhysics_js_function_and_add_to_event(cls):
        return ["var isPhysicsOn = true;",
                "function disablePhysics()",
                "{",
                "   //use this if the layout IS hierarchical",
                "   if(isPhysicsOn && network['layoutEngine']['options']['hierarchical']['enabled'])",
                "   {",
                # '       alert("disabled by disablePhysics")' ,
                '       newOptions={"layout":{"hierarchical":{"enabled":false}}};\n',
                "       network.setOptions(newOptions);",
                '       newOptions={\"physics\":{\"enabled\":false}};\n',
                "       network.setOptions(newOptions);",
                "       isPhysicsOn = false;\n",
                "   }",
                "};",

                "function onStabilized()",
                "{",
                "   //use this if the layout is NOT hierarchical",
                "   if(isPhysicsOn && !(network['layoutEngine']['options']['hierarchical']['enabled']))",
                "   {",
                # '       alert("disabled by stabilized")' ,
                '       newOptions={"layout":{"hierarchical":{"enabled":false}}};\n',
                "       network.setOptions(newOptions);",
                '       newOptions={\"physics\":{\"enabled\":false}};\n',
                "       network.setOptions(newOptions);",
                # "       sendCurrentCombinedDataToPython()",
                "       isPhysicsOn = false;\n",
                "   }",
                "};",
                "function onIterationStabilized()",
                "{",
                "   //use this if the layout is NOT hierarchical",
                "   if(isPhysicsOn && !(network['layoutEngine']['options']['hierarchical']['enabled']))",
                "   {",
                # '       alert("disabled by iteration stabilized")',
                '       newOptions={"layout":{"hierarchical":{"enabled":false}}};\n',
                "       network.setOptions(newOptions);",
                '       newOptions={\"physics\":{\"enabled\":false}};\n',
                "       network.setOptions(newOptions);",
                # "       sendCurrentCombinedDataToPython()",
                "       isPhysicsOn = false;\n",
                "   }",
                "};",

                ]

    @classmethod
    def create_AddEdge_js_function(cls):
        return ["function AddEdge(id,from_id, to_id,color,arrow,hidden)",
                "{",
                '   applyAddEdgesToNetwork([{'
                '       "id": id,'
                '       "arrowStrikethrough": false,'
                '       "arrows": arrow,'
                '       "color": color,'
                '       "from": from_id,'
                '       "hidden": hidden,'
                '       "smooth": '
                '           {'
                '           "enabled": false'
                '           },'
                '               "to": to_id,'
                '               "width": 2'
                '           }]);',
                "}"]

    @classmethod
    def create_sendCurrentCombinedDataToPython_js_function(cls):
        return ["function sendCurrentCombinedDataToPython()",
                "{",
                # '   unhideAllRelations()',
                '   network.storePositions() // alters the data in network.body.data.nodes with the current coordinates so i can be read and stored',
                # '   console.log("called storePositions()")',
                "   var nodeList = ExtractNodeList();    //nodes and their position (including edgeJointNodes)",
                "   var edgeList = ExtractEdgeList();    //edges (including subEdges)",
                "   ",
                "   var relationIdToSubEdgesList = Array.from(relationIdToSubEdges.entries());   //data supporting dynamic removal functionality",
                "   var relationIdToTotalSubEdgeCountList = Array.from(relationIdToTotalSubEdgeCount.entries());   //data supporting dynamic removal functionality",
                "   var relationIdToJointNodesList = Array.from(relationIdToJointNodes.entries());   //data supporting dynamic removal functionality",
                "   var SubEdgesToOriginalRelationIdList = Array.from(SubEdgesToOriginalRelationId.entries());   //data supporting dynamic removal functionality",
                "   var edgeJointNodesIdToConnectionDataDictList = Array.from(edgeJointNodesIdToConnectionDataDict.entries());   //data supporting right-click edgeJointNode removal functionality",
                "   ",
                "   //set all relations to unhidden if they were hidden by the python interface",
                "   setAllNonCollectionEdgesToUnhidden(edgeList);",
                "   ",
                "   ",
                # "   var edgeJointNodesIdToConnectionDataDictList = JSON.parse(edgeJointNodesIdToConnectionDataDict;   //data supporting right-click edgeJointNode removal functionality",
                "   var combinedData = {'html_path': html_path,"
                "                       'nodeList': nodeList,"
                "                       'edgeList': edgeList,"
                "                       'relationIdToSubEdgesList':relationIdToSubEdgesList,"
                "                       'relationIdToTotalSubEdgeCountList': relationIdToTotalSubEdgeCountList,"
                "                       'relationIdToJointNodesList': relationIdToJointNodesList,"
                "                       'SubEdgesToOriginalRelationIdList': SubEdgesToOriginalRelationIdList,"
                "                       'edgeJointNodesIdToConnectionDataDictList' : edgeJointNodesIdToConnectionDataDictList,"
                "                       'collection_id_to_list_of_relation_ids' : collection_id_to_list_of_relation_ids"
                "                       }",
                '   var combinedDataStr = JSON.stringify(combinedData)',
                # '   console.log(combinedDataStr)',
                # '   alert("DataVisualisationScreen: " + combinedDataStr);'
                "   if (window.backend)",
                "   {",
                "       window.backend.receive_new_combined_data(combinedDataStr);",
                # "       alert('sent new node'); ",
                # "       window.backend.receive_coordinates(JSON.stringify({lat: 56, lng: 30}));",
                "   }"
                "   else",
                "   {"
                '       console.log("QWebChannel is not initialized yet.");',
                # '       alert("DataVisualisationScreen: QWebChannel is not initialized");',
                "   }",

                "}"]

    @classmethod
    def create_ExtractNodeList_js_function(cls):
        return ["function ExtractNodeList()   "
                "{"
                "   var node_attributes = Object.fromEntries(network.body.data.nodes._data);",
                "   var nodeList = [];",
                "   for (const nodeId of network.body.data.nodes._data.keys()) ",
                "   {",
                "       if (network.body.data.nodes._data.get(nodeId).title)",
                "       {",
                "           var strTitle = network.body.data.nodes._data.get(nodeId).title.innerHTML; // use the innerhtml so it can be converted to json",
                "           if (strTitle)",
                "           node_attributes[nodeId].title = 'htmlTitle(\"' + strTitle.replace('\"','\\\"') +  '\")'; //add htmlTitle so is looks exactly like how the nodes are normally created",
                "       }",
                "       nodeList.push(node_attributes[nodeId]);",
                "   }",
                "   return nodeList",
                "}"]

    @classmethod
    def create_setAllNonCollectionEdgesToUnhidden_js_function(cls):
        return ["function setAllNonCollectionEdgesToUnhidden(edgeList)",
                '{',
                '   //collect all relation_ids that are in a collection they should not be unhidden',
                '   var relation_ids_in_collection_id = getFlatListOfRelationIdsInCollection()',
                '   edgeList.forEach((edgeData) =>',
                '   {',
                '       if( !relation_ids_in_collection_id.includes(edgeData["id"]))',
                '           edgeData["hidden"] = false;',
                '   })',
                '   return edgeList',
                '} ']

    @classmethod
    def create_getFlatListOfRelationIdsInCollection(cls):
        return ["function getFlatListOfRelationIdsInCollection()",
                '{',
                '   relation_ids_in_collection_id = []',
                '   for (collection_id in collection_id_to_list_of_relation_ids)',
                '   {',
                '       for (rel_tuple_index in collection_id_to_list_of_relation_ids[collection_id])',
                '       {',
                '           relation_ids_in_collection_id.push(collection_id_to_list_of_relation_ids[collection_id][rel_tuple_index][0])',
                '       }',
                '   }',
                '   return relation_ids_in_collection_id;',
                '}']

    @classmethod
    def create_ExtractEdgeList_js_function(cls):
        return ["function ExtractEdgeList()   "
                "{"
                "   var edge_attributes = Object.fromEntries(network.body.data.edges._data);",
                "   var edgeList = [];",
                "   for (const edgeId of network.body.data.edges._data.keys()) ",
                "   {",
                "       edgeList.push(edge_attributes[edgeId]);",
                "   }",
                "   return edgeList",
                "}"]
    @classmethod
    def create_AddEdgeWithLabel_js_function(cls):
        return ["function AddEdgeWithLabel(id,from_id, to_id,color,arrow,label,hidden)",
                "{",
                '   applyAddEdgesToNetwork([{'
                '   "id": id,'
                '   "arrowStrikethrough": false,'
                '   "arrows": arrow,'
                '   "color": color,'
                '   "from": from_id,'
                '   "label": label,'
                '   "smooth": {'
                '   "enabled": false'
                '    },'
                '   "hidden": hidden,'
                '   "to": to_id,'
                '   "width": 2'
                '    }]);',
                "}"]



    @classmethod
    def create_removeEdge_js_function(cls):
        return ['function removeEdge(id)',
                '{',
                '  if (network.body.data.edges._data.has(id))',
                '       applyRemoveEdgesFromNetwork([id]);//defined in PyViswrapper',
                '  else if (!relationIdToSubEdges)',
                '      console.log("attempted to remove: " + id)',
                '  else if (relationIdToSubEdges.has(id))',
                '  {',
                '      //remove all selfmade subEdges and jointNodes that the original relations was replaced with',
                '      subEdges = relationIdToSubEdges.get(id)',
                '      jointNodes = relationIdToJointNodes.get(id)',
                '      ',
                '      applyRemoveEdgesFromNetwork(subEdges); //defined in PyViswrapper',
                '      applyRemoveNodesFromNetwork(jointNodes); //defined in PyViswrapper',
                '      ',
                '      //remove stored data on subedges and jointnodes',
                '      subEdges.forEach((subEdgeId) =>',
                '      {',
                '          SubEdgesToOriginalRelationId.delete(subEdgeId);',
                '      })',
                '      relationIdToTotalSubEdgeCount.delete(id);',
                '      relationIdToSubEdges.delete(id);',
                '      relationIdToJointNodes.delete(id);',
                '  }',
                '  else',
                '      {console.log("attempted to remove: " + id)}',
                '}']

    @classmethod
    def create_UpdateCollectionAttributes_js_function(cls):
        return ['function UpdateCollectionAttributes(collection_id,new_label,new_title,new_collection_id_to_list_of_relation_ids)',
                '{',
                "      var newPos = network.getPosition(collection_id)",
                '      applyUpdateNodeInNetwork({"id": collection_id, "label": new_label, "title":new_title,"x":  newPos.x,"y": newPos.y});',
                '      collection_id_to_list_of_relation_ids = new_collection_id_to_list_of_relation_ids;',
                '}']

    @classmethod
    def create_UpdateAllRelationHiddenStatesOfColor_js_function(cls):
        return [
            'function UpdateAllRelationHiddenStatesOfColor(color,hidden)',
            '{',
            '       var newAttributesEdgesList = [];'
            '       network.body.data.edges._data.forEach((data,key) =>',
            '       {',
            '           var relation_ids_in_collection_id = getFlatListOfRelationIdsInCollection()',
            # '           if (data["color"] == color &&(data["id"].includes("special_edge")|| !relation_ids_in_collection_id.includes(data["id"])))'  ,
            '           if (data["color"] == color && !relation_ids_in_collection_id.includes(data["id"]))',
            '               newAttributesEdgesList.push({"id": data["id"], "hidden": hidden});',
            '       })',
            '       applyUpdateEdgeInNetwork(newAttributesEdgesList,false);',
            '}'
        ]

    @classmethod
    def create_unhideAllRelations_js_function(cls):
        return [
            'function unhideAllRelations()',
            '{',
            '       var newAttributesEdgesList = [];'
            '       network.body.data.edges._data.forEach((data,key) =>',
            '       {',
            '               newAttributesEdgesList.push({"id": data["id"], "hidden": false});'
            '       })',
            '       applyUpdateEdgeInNetwork(newAttributesEdgesList,false);',
            '}'
        ]

    @classmethod
    def create_removeEdgeJointNode_js_function(cls):
        return ['function removeEdgeJointNode(node_id)',
                '{',
                '   if (edgeJointNodesIdToConnectionDataDict.has(node_id))',
                '   {',
                '       var edgeNodeConnectionData = edgeJointNodesIdToConnectionDataDict.get(node_id);',
                '       var originalEdgeId = edgeNodeConnectionData["originalEdgeId"]; ',
                '       var previousEdgeId = edgeNodeConnectionData["previousEdgeId"];',
                '       var subEdge1Id = edgeNodeConnectionData["newSubEdge1Id"]; ',
                '       var subEdge2Id = edgeNodeConnectionData["newSubEdge2Id"]; ',
                '       var toNodeId = edgeNodeConnectionData["newSubEdge1Data.to"]; ',
                '       var fromNodeId = edgeNodeConnectionData["newSubEdge2Data.from"]; ',
                '       ',
                '       //get attribute of a subedge',
                '       newEdgeAttributes = JSON.parse(JSON.stringify(network.body.data.edges._data.get(subEdge1Id)));',
                '       //remove the 2 subedges in network',
                '       applyRemoveEdgesFromNetwork([subEdge1Id,subEdge2Id]);',
                '       ',
                '       //also remove them from supporting data'
                '       SubEdgesToOriginalRelationId.delete(subEdge1Id);',
                '       SubEdgesToOriginalRelationId.delete(subEdge2Id);',

                '       var currCount = relationIdToTotalSubEdgeCount.get(originalEdgeId);',
                '       relationIdToTotalSubEdgeCount.set(originalEdgeId, currCount-1);',
                '       relationIdToSubEdges.get(originalEdgeId).splice(subEdge1Id,1);',
                '       relationIdToSubEdges.get(originalEdgeId).splice(subEdge2Id,1);',
                '       relationIdToSubEdges.get(originalEdgeId).push(previousEdgeId)',
                '       SubEdgesToOriginalRelationId.set(previousEdgeId,originalEdgeId);',

                '       //add a new subedge with originalEdgeId between the to and from node',
                '       newEdgeAttributes.id = previousEdgeId;'   ,
                '       newEdgeAttributes.to = toNodeId;',
                '       newEdgeAttributes.from = fromNodeId; ',
                "       applyAddEdgesToNetwork([newEdgeAttributes])",
                '       ',
                '       //edit supporting data of the to and from nodes so they have the correct',
                '       //to and from themselves',
                '       updateNeighbouringEdgeJointNode(toNodeId, node_id, fromNodeId)',
                '       updateNeighbouringEdgeJointNode(fromNodeId, node_id, toNodeId)',
                '       updateConnectingEdgeOnNeighbouringEdgeJointNode(toNodeId,subEdge1Id,previousEdgeId)',
                '       updateConnectingEdgeOnNeighbouringEdgeJointNode(toNodeId,subEdge2Id,previousEdgeId)',
                '       updateConnectingEdgeOnNeighbouringEdgeJointNode(fromNodeId,subEdge1Id,previousEdgeId)',
                '       updateConnectingEdgeOnNeighbouringEdgeJointNode(fromNodeId,subEdge2Id,previousEdgeId)',
                '       ' ,
                # '   console.log("remove edge jointnode")',
                # 'edgeJointNodesIdToConnectionDataDict.forEach((data,key) =>',
                # '      {',
                # '         console.log(key +": " + JSON.stringify(data))',
                # '      })',
                # '   console.log(" network.body.data.edges:\\n ")',
                # 'network.body.data.edges._data.forEach((data,key) =>',
                # '      {',
                # '         console.log(key +": " + JSON.stringify(data))',
                # '      })',
                 # '   console.log(" edges:\\n "+  Array.from(network.body.data.edges._data.entries()))',
                '       //remove edgeJointNode from supporting data',
                '       relationIdToJointNodes.delete(node_id);',
                '       //remove edgeJointNode from network',
                '       applyRemoveNodesFromNetwork([node_id])',
                '       currentlyHoveredNode = null',
                '   }',
                '}',

                ]

    @classmethod
    def remove_relations(cls,to_remove_list, vis_wrap, webview):
        if not vis_wrap:
            return

        # if there are removed relations remove them from the visualisation
        for relation_object in to_remove_list:
            rel_id = relation_object.assetId.identificator

            if rel_id in vis_wrap.relation_id_to_collection_id:
                collection_ids = vis_wrap.relation_id_to_collection_id.pop(rel_id)
                js_code = ""
                for collection_id in collection_ids:
                    vis_wrap.collection_id_to_list_of_relation_ids[collection_id] = \
                        [rel_set for rel_set in vis_wrap.collection_id_to_list_of_relation_ids[
                            collection_id] if rel_set[0] != rel_id]
                    new_label, new_title = cls.create_new_special_node_text(collection_id,vis_wrap)

                    js_code += f'UpdateCollectionAttributes("{collection_id}", "{new_label}","{new_title}",{json.dumps(vis_wrap.collection_id_to_list_of_relation_ids)});\n'
                js_code += f'\nremoveEdge("{relation_object.assetId.identificator}");'
            else:
                js_code = f'removeEdge("{relation_object.assetId.identificator}");'

            OTLLogger.logger.debug(js_code)
            webview.page().runJavaScript(js_code)

    @classmethod
    def add_new_relations(cls, to_add_list, vis_wrap,webview,relation_visible_dict):

        if not vis_wrap:
            return
        # if there are new relations add them to the visualisation
        for relation_object in to_add_list:
            rel_id = relation_object.assetId.identificator
            add_edge_arguments = vis_wrap.create_edge_inject_arguments(relation_object)
            collect_id_to_original_asset_id_dict= {}

            if "label" in add_edge_arguments:  # a heeftBetrokkene relation with their rol as label
                # first check if the new relation needs to be added to a current collection
                added_to_collection = False
                for special_edge in vis_wrap.special_edges:
                    # first check if the new relation is the same type as the special edge
                    if (special_edge["label"] == add_edge_arguments["label"] and
                            special_edge["arrows"] == add_edge_arguments["arrow"] and
                            special_edge["color"] == add_edge_arguments["color"]):
                        if special_edge["from"] == add_edge_arguments["from_id"]:
                            # is this case:
                            # special_edge["from"] is the id of the asset that has relations to many assets
                            collection_id = special_edge["to"]  # special_edge["to"] is the id of the collection_node
                            try:
                                screen_name_of_new_asset = vis_wrap.asset_id_to_display_name_dict[add_edge_arguments["to_id"]]
                            except:
                                OTLLogger.logger.debug(f"Couldn't find display name of node id: {add_edge_arguments['to_id']}\n Needs refresh")
                                screen_name_of_new_asset = vis_wrap.asset_id_to_display_name_dict[collect_id_to_original_asset_id_dict[add_edge_arguments["to_id"]]]

                            added_to_collection = True
                            js_code = cls.create_js_code_to_add_to_collection(vis_wrap,
                                                                               collection_id,
                                                                               rel_id,
                                                                               screen_name_of_new_asset)
                            OTLLogger.logger.debug(js_code)
                            webview.page().runJavaScript(js_code)
                            # point the original relation to the collection instead of its intended target
                            collect_id_to_original_asset_id_dict[collection_id] = add_edge_arguments["from_id"]
                            add_edge_arguments["from_id"] = collection_id

                        elif special_edge["to"] == add_edge_arguments["to_id"]:
                            # is this case:
                            # special_edge["to"] is the id of the asset that has relations to many assets
                            collection_id = special_edge["from"]  # special_edge["from"] is the id of the collection_node
                            try:
                                screen_name_of_new_asset = vis_wrap.asset_id_to_display_name_dict[add_edge_arguments["from_id"]]
                            except:
                                OTLLogger.logger.debug(f"Couldn't find display name of node id: {add_edge_arguments['from_id']}\n Needs refresh")
                                screen_name_of_new_asset = vis_wrap.asset_id_to_display_name_dict[collect_id_to_original_asset_id_dict[add_edge_arguments["from_id"]]]
                            added_to_collection = True
                            js_code = cls.create_js_code_to_add_to_collection(vis_wrap,
                                                                               collection_id,
                                                                               rel_id,
                                                                               screen_name_of_new_asset)
                            OTLLogger.logger.debug(js_code)
                            webview.page().runJavaScript(js_code)
                            # point the original relation to the collection instead of its intended target
                            collect_id_to_original_asset_id_dict[collection_id] = \
                            add_edge_arguments["to_id"]
                            add_edge_arguments["to_id"] = collection_id

                if not added_to_collection:
                    js_bool_hidden = "false"
                    rel_type = relation_object.typeURI
                    if rel_type in relation_visible_dict.keys() and not relation_visible_dict[rel_type]:
                        js_bool_hidden = "true"

                    js_code = f'AddEdgeWithLabel("{add_edge_arguments["id"]}","{add_edge_arguments["from_id"]}", "{add_edge_arguments["to_id"]}","{add_edge_arguments["color"]}","{add_edge_arguments["arrow"]}","{add_edge_arguments["label"]}", {js_bool_hidden})'
                    OTLLogger.logger.debug(js_code)
                    webview.page().runJavaScript(js_code)
                else:
                    js_code = f'\nAddEdgeWithLabel("{add_edge_arguments["id"]}","{add_edge_arguments["from_id"]}", "{add_edge_arguments["to_id"]}","{add_edge_arguments["color"]}","{add_edge_arguments["arrow"]}","{add_edge_arguments["label"]}", true)'
                    OTLLogger.logger.debug(js_code)
                    webview.page().runJavaScript(js_code)

            elif "arrow" in add_edge_arguments:  # a directional relation
                # first check if the new relation needs to be added to a current collection
                added_to_collection = False
                for special_edge in vis_wrap.special_edges:
                    # first check if the new relation is the same type as the special edge
                    if (special_edge["arrows"] == add_edge_arguments["arrow"] and
                            special_edge["color"] == add_edge_arguments["color"]):
                        if special_edge["from"] == add_edge_arguments["from_id"]:
                            # is this case:
                            # special_edge["from"] is the id of the asset that has relations to many assets
                            collection_id = special_edge[
                                "to"]  # special_edge["to"] is the id of the collection_node
                            try:
                                screen_name_of_new_asset = \
                                    vis_wrap.asset_id_to_display_name_dict[
                                        add_edge_arguments["to_id"]]
                            except:
                                OTLLogger.logger.debug(f"Couldn't find display name of node id: {add_edge_arguments['to_id']}\n Needs refresh")
                                screen_name_of_new_asset = vis_wrap.asset_id_to_display_name_dict[collect_id_to_original_asset_id_dict[add_edge_arguments["to_id"]]]


                            added_to_collection = True
                            js_code = cls.create_js_code_to_add_to_collection(vis_wrap,
                                                                               collection_id,
                                                                               rel_id,
                                                                               screen_name_of_new_asset)
                            OTLLogger.logger.debug(js_code)
                            webview.page().runJavaScript(js_code)
                            # point the original relation to the collection instead of its intended target
                            collect_id_to_original_asset_id_dict[collection_id] = \
                                add_edge_arguments["to_id"]
                            add_edge_arguments["from_id"] = collection_id
                        elif special_edge["to"] == add_edge_arguments["to_id"]:
                            # is this case:
                            # special_edge["to"] is the id of the asset that has relations to many assets
                            collection_id = special_edge[
                                "from"]  # special_edge["from"] is the id of the collection_node
                            try:
                                screen_name_of_new_asset = \
                                    vis_wrap.asset_id_to_display_name_dict[
                                        add_edge_arguments["from_id"]]
                            except:
                                OTLLogger.logger.debug(
                                f"Couldn't find display name of node id: {add_edge_arguments['from_id']}\n Needs refresh")
                                screen_name_of_new_asset = vis_wrap.asset_id_to_display_name_dict[collect_id_to_original_asset_id_dict[add_edge_arguments["from_id"]]]
                            added_to_collection = True
                            js_code = cls.create_js_code_to_add_to_collection(vis_wrap,
                                                                               collection_id,
                                                                               rel_id,
                                                                               screen_name_of_new_asset)
                            OTLLogger.logger.debug(js_code)
                            webview.page().runJavaScript(js_code)
                            # point the original relation to the collection instead of its intended target
                            collect_id_to_original_asset_id_dict[collection_id] = \
                                add_edge_arguments["to_id"]
                            add_edge_arguments["to_id"] = collection_id

                if not added_to_collection:
                    js_bool_hidden = "false"
                    rel_type = relation_object.typeURI
                    if rel_type in relation_visible_dict.keys() and not relation_visible_dict[
                        rel_type]:
                        js_bool_hidden = "true"
                    js_code = f'AddEdge("{add_edge_arguments["id"]}","{add_edge_arguments["from_id"]}", "{add_edge_arguments["to_id"]}","{add_edge_arguments["color"]}","{add_edge_arguments["arrow"]}", {js_bool_hidden});'
                    OTLLogger.logger.debug(js_code)
                    webview.page().runJavaScript(js_code)
                else:
                    js_code = f'AddEdge("{add_edge_arguments["id"]}","{add_edge_arguments["from_id"]}", "{add_edge_arguments["to_id"]}","{add_edge_arguments["color"]}","{add_edge_arguments["arrow"]}", true);'
                    OTLLogger.logger.debug(js_code)
                    webview.page().runJavaScript(js_code)
            else:  # a bidirectional relation
                # first check if the new relation needs to be added to a current collection
                added_to_collection = False

                for special_edge in vis_wrap.special_edges:
                    # first check if the new relation is the same type as the special edge
                    if (special_edge["color"] == add_edge_arguments["color"]):
                        if (special_edge["from"] == add_edge_arguments["from_id"] or
                                special_edge["from"] == add_edge_arguments["to_id"]):
                            # is this case:
                            # special_edge["from"] is the id of the asset that has relations to many assets
                            collection_id = special_edge["to"]  # special_edge["to"] is the id of the collection_node

                            if special_edge["from"] == add_edge_arguments["to_id"]:
                                try:
                                    screen_name_of_new_asset = \
                                        vis_wrap.asset_id_to_display_name_dict[
                                            add_edge_arguments["from_id"]]
                                except:
                                    OTLLogger.logger.debug(
                                        f"Couldn't find display name of node id: {add_edge_arguments['from_id']}\n Needs refresh")
                                    OTLLogger.logger.debug(
                                        f"add_edge_arguments[\"from_id\"]: {add_edge_arguments['from_id']}\ncollect_id_to_original_asset_id_dict: {collect_id_to_original_asset_id_dict}\n  vis_wrap.asset_id_to_display_name_dict: {vis_wrap.asset_id_to_display_name_dict}")

                                    screen_name_of_new_asset = \
                                    vis_wrap.asset_id_to_display_name_dict[
                                        collect_id_to_original_asset_id_dict[add_edge_arguments["from_id"]]]
                            elif special_edge["from"] == add_edge_arguments["from_id"]:
                                try:
                                    screen_name_of_new_asset = \
                                        vis_wrap.asset_id_to_display_name_dict[
                                            add_edge_arguments["to_id"]]

                                except:
                                    OTLLogger.logger.debug(
                                    f"Couldn't find display name of node id: {add_edge_arguments['to_id']}\n Needs refresh")
                                    OTLLogger.logger.debug(
                                        f"add_edge_arguments[\"to_id\"]: {add_edge_arguments['to_id']}\ncollect_id_to_original_asset_id_dict: {collect_id_to_original_asset_id_dict}\n  vis_wrap.asset_id_to_display_name_dict: {vis_wrap.asset_id_to_display_name_dict}")

                                    screen_name_of_new_asset = vis_wrap.asset_id_to_display_name_dict[collect_id_to_original_asset_id_dict[add_edge_arguments["to_id"]]]
                            added_to_collection = True
                            js_code =  cls.create_js_code_to_add_to_collection(vis_wrap,
                                                                               collection_id,
                                                                               rel_id,
                                                                               screen_name_of_new_asset)
                            OTLLogger.logger.debug(js_code)
                            webview.page().runJavaScript(js_code)
                            # point the original relation to the collection instead of its intended target
                            # collect_id_to_original_asset_id_dict[collection_id] = \
                            #     special_edge["from"]
                            if special_edge["from"] == add_edge_arguments["to_id"]:
                                add_edge_arguments["to_id"] = collection_id
                                collect_id_to_original_asset_id_dict[collection_id] = \
                                    add_edge_arguments["to_id"]
                            elif special_edge["from"] == add_edge_arguments["from_id"]:
                                add_edge_arguments["from_id"] = collection_id
                                collect_id_to_original_asset_id_dict[collection_id] = \
                                    add_edge_arguments["from_id"]

                        elif special_edge["to"] == add_edge_arguments["to_id"] or special_edge["to"] == add_edge_arguments["from_id"]:
                            # is this case:
                            # special_edge["to"] is the id of the asset that has relations to many assets
                            collection_id = special_edge["from"]  # special_edge["from"] is the id of the collection_node

                            if special_edge["to"] == add_edge_arguments["to_id"]:
                                try:
                                    screen_name_of_new_asset = \
                                        vis_wrap.asset_id_to_display_name_dict[
                                            add_edge_arguments["from_id"]]
                                except:
                                    OTLLogger.logger.debug(
                                        f"Couldn't find display name of node id: {add_edge_arguments['from_id']}\n Needs refresh")
                                    OTLLogger.logger.debug(
                                        f"add_edge_arguments[\"from_id\"]: {add_edge_arguments['from_id']}\ncollect_id_to_original_asset_id_dict: {collect_id_to_original_asset_id_dict}\n  vis_wrap.asset_id_to_display_name_dict: {vis_wrap.asset_id_to_display_name_dict}")
                                    screen_name_of_new_asset = \
                                    vis_wrap.asset_id_to_display_name_dict[
                                        collect_id_to_original_asset_id_dict[
                                            add_edge_arguments["from_id"]]]

                            elif special_edge["to"] == add_edge_arguments["from_id"]:
                                try:
                                    screen_name_of_new_asset = \
                                        vis_wrap.asset_id_to_display_name_dict[
                                            add_edge_arguments["to_id"]]
                                except:
                                    OTLLogger.logger.debug(
                                        f"Couldn't find display name of node id: {add_edge_arguments['to_id']}\n Needs refresh")
                                    OTLLogger.logger.debug(
                                        f"add_edge_arguments[\"to_id\"]: {add_edge_arguments['to_id']}\ncollect_id_to_original_asset_id_dict: {collect_id_to_original_asset_id_dict}\n  vis_wrap.asset_id_to_display_name_dict: {vis_wrap.asset_id_to_display_name_dict}")
                                    screen_name_of_new_asset = \
                                    vis_wrap.asset_id_to_display_name_dict[
                                        collect_id_to_original_asset_id_dict[
                                            add_edge_arguments["to_id"]]]



                            added_to_collection = True
                            js_code = cls.create_js_code_to_add_to_collection(vis_wrap,
                                                                               collection_id,
                                                                               rel_id,
                                                                               screen_name_of_new_asset)
                            OTLLogger.logger.debug(js_code)
                            webview.page().runJavaScript(js_code)
                            # point the original relation to the collection instead of its intended target
                            # collect_id_to_original_asset_id_dict[collection_id] = \
                            #     special_edge["to"]
                            if special_edge["to"] == add_edge_arguments["to_id"]:
                                collect_id_to_original_asset_id_dict[collection_id] = \
                                    add_edge_arguments["to_id"]
                                add_edge_arguments["to_id"] = collection_id
                            elif special_edge["to"] == add_edge_arguments["from_id"]:
                                collect_id_to_original_asset_id_dict[collection_id] = \
                                    add_edge_arguments["from_id"]
                                add_edge_arguments["from_id"] = collection_id


                if not added_to_collection:
                    js_bool_hidden = "false"
                    rel_type = relation_object.typeURI
                    if rel_type in relation_visible_dict.keys() and not relation_visible_dict[
                        rel_type]:
                        js_bool_hidden = "true"
                    js_code = f'AddEdge("{add_edge_arguments["id"]}","{add_edge_arguments["from_id"]}", "{add_edge_arguments["to_id"]}","{add_edge_arguments["color"]}",null, {js_bool_hidden});'
                    OTLLogger.logger.debug(js_code)
                    webview.page().runJavaScript(js_code)
                else:
                    js_code = f'\nAddEdge("{add_edge_arguments["id"]}","{add_edge_arguments["from_id"]}", "{add_edge_arguments["to_id"]}","{add_edge_arguments["color"]}",null, true);'
                    OTLLogger.logger.debug(js_code)
                    webview.page().runJavaScript(js_code)


    @classmethod
    def create_js_code_to_add_to_collection(cls, vis_wrap, collection_id, rel_id,
                                            screen_name_of_new_asset):
        new_rel_id_and_asset_screen_name_set = (rel_id, screen_name_of_new_asset)
        vis_wrap.relation_id_to_collection_id[rel_id].append(collection_id)
        vis_wrap.collection_id_to_list_of_relation_ids[
            collection_id].append(new_rel_id_and_asset_screen_name_set)
        new_label, new_title = cls.create_new_special_node_text(collection_id, vis_wrap)

        js_code = f'UpdateCollectionAttributes("{collection_id}", "{new_label}","{new_title}",{json.dumps(vis_wrap.collection_id_to_list_of_relation_ids)});'

        return js_code

    @classmethod
    def create_new_special_node_text(cls, collection_id, vis_wrap):
        new_collection_content = "\n".join([rel_set[1] for rel_set in
                                            vis_wrap.collection_id_to_list_of_relation_ids[
                                                collection_id]])
        new_collection_size = len(
            vis_wrap.collection_id_to_list_of_relation_ids[
                collection_id])
        new_label = f"<i><b>Collectie ({new_collection_size})</b></i>"
        new_title = f"Collectie ({new_collection_size}):\n{new_collection_content}"
        # next line is so it can be communicated the to javascript correctly
        new_title = new_title.replace("\n", "\\n")
        return new_label, new_title



from pathlib import Path
from typing import Callable

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QDialog, QLabel, QVBoxLayout, QHBoxLayout, QLineEdit, QPushButton, QDialogButtonBox

from otlmow_gui.Domain import global_vars
from otlmow_gui.Domain.project.Project import Project
from otlmow_gui.Domain.step_domain.HomeDomain import HomeDomain

import qtawesome as qta

from otlmow_gui.Exceptions.EmptyFieldError import EmptyFieldError
from otlmow_gui.Exceptions.WrongDatabaseError import WrongDatabaseError
from otlmow_gui.GUI.dialog_windows.NotificationWindow import NotificationWindow
from otlmow_gui.GUI.dialog_windows.ProjectExistsError import ProjectExistsError
from otlmow_gui.GUI.dialog_windows.file_picker_dialog.SubsetLoadFilePickerDialog import \
    SubsetLoadFilePickerDialog


class UpsertProjectWindow(QDialog):

    def __init__(self, language_settings: Callable, project: Project = None):
        super().__init__()

        self._ = language_settings

        self.error_label = QLabel()
        self.error_label.setText("")

        self.setMinimumWidth(400)
        # Makes the dialog the primary screen, disabling the screen behind it
        self.setModal(True)

        # Creates the vertical stack layout
        layout = QVBoxLayout()

        # Creates the input fields
        container_eigen_ref, input_eigen_ref = self.create_eigen_ref_container()
        container_bestek, input_bestek = self.create_bestek_container()
        container_subset, input_subset = self.create_subset_container()

        # Adds the input fields to the layout
        layout.addLayout(container_eigen_ref)
        layout.addLayout(container_bestek)
        layout.addLayout(container_subset)

        if project:
            self.setWindowTitle(self._("alter_project_title"))
            input_eigen_ref.setText(project.eigen_referentie)
            input_bestek.setText(project.bestek)
            input_subset.setText(str(project.subset_path))
        else:
            self.setWindowTitle(self._("new_project_title"))

        # Changes the color of the error label to red
        self.error_label.setStyleSheet("color: red")

        # Creates the button box
        button_box = self.create_button_box()
        # sends the values off to validate once submitted
        button_box.accepted.connect(
            lambda: self.pass_values_through_validate(input_eigen_ref= input_eigen_ref.text(),
                                                      input_bestek= input_bestek.text(),
                                                      input_subset= input_subset.text(),
                                                      project=project))
        button_box.rejected.connect(self.reject)
        # Adds the two buttons to the layout
        layout.addWidget(button_box)
        layout.addWidget(self.error_label)
        # Fills the dialog with the created layout
        self.setLayout(layout)

    def create_subset_container(self) -> tuple[QHBoxLayout,QLineEdit]:
        """
        Creates a container for the subset input field along with its label and file picker button.

        This method constructs a horizontal layout that includes a label indicating the subset, a read-only input field for displaying the subset value, and a button that allows the user to open a file picker dialog. The layout and input field are returned for further use in the user interface.

        :return: A tuple containing the layout for the subset container and the input field.
        :rtype: tuple[QHBoxLayout, QLineEdit]
        """

        label_subset = QLabel(self._("subset") + ":")
        input_subset = QLineEdit()
        input_subset.setObjectName("subset_edit")
        input_subset.setReadOnly(True)
        input_subset.setPlaceholderText(self._("subset"))
        file_picker_btn = QPushButton()
        file_picker_btn.setObjectName("file_picker_btn")
        file_picker_btn.setIcon(qta.icon('mdi.folder-open-outline'))
        file_picker_btn.clicked.connect(lambda: self.open_file_picker(input_subset))
        container_subset = QHBoxLayout()
        container_subset.addWidget(label_subset, alignment=Qt.AlignmentFlag.AlignLeft)
        container_subset.addWidget(input_subset)
        container_subset.addWidget(file_picker_btn)
        return container_subset, input_subset

    def create_bestek_container(self) -> tuple[QHBoxLayout,QLineEdit]:
        """
        Creates a container for the bestek input field along with its label.

        This method constructs a horizontal layout that includes a label indicating the service order and an input field for entering the bestek value. The layout and input field are returned for further use in the user interface.

        :return: A tuple containing the layout for the bestek container and the input field.
        :rtype: tuple[QHBoxLayout, QLineEdit]
        """

        label_bestek = QLabel(self._("service_order") + ":")
        input_bestek = QLineEdit()
        input_bestek.setObjectName("bestek_edit")
        input_bestek.setPlaceholderText(self._("service_order"))
        container_bestek = QHBoxLayout()
        container_bestek.addWidget(label_bestek, alignment=Qt.AlignmentFlag.AlignLeft)
        container_bestek.addWidget(input_bestek)
        return container_bestek, input_bestek

    def create_eigen_ref_container(self)  -> tuple[QHBoxLayout,QLineEdit]:
        """
        Creates a container for the eigen reference input field along with its label.

        This method constructs a horizontal layout that includes a label indicating the own reference and an input field for entering the eigen reference value. The layout and input field are returned for further use in the user interface.

        :return: A tuple containing the layout for the eigen reference container and the input field.
        :rtype: tuple[QHBoxLayout, QLineEdit]
        """

        label_eigen_ref = QLabel(self._("own_reference") + ":")
        input_eigen_ref = QLineEdit()
        input_eigen_ref.setObjectName("eigen_ref_edit")
        input_eigen_ref.setPlaceholderText(self._("own_reference"))
        container_eigen_ref = QHBoxLayout()
        container_eigen_ref.addWidget(label_eigen_ref, alignment=Qt.AlignmentFlag.AlignLeft)
        container_eigen_ref.addWidget(input_eigen_ref)
        return container_eigen_ref, input_eigen_ref

    def create_button_box(self) -> QDialogButtonBox:
        """
        Creates a button box for the dialog window with OK and Cancel buttons.

        This method initializes a QDialogButtonBox with standard OK and Cancel buttons, applying specific properties and text labels to each button. The button box is configured for use in the dialog interface and returned for further integration.

        :return: The configured QDialogButtonBox containing the OK and Cancel buttons.
        :rtype: QDialogButtonBox
        """

        button_box = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)
        button_box.setProperty("class", "button-box")
        button_box.button(QDialogButtonBox.StandardButton.Ok).setProperty("class", "primary-button")
        button_box.button(QDialogButtonBox.StandardButton.Ok).setText(self._("submit"))
        button_box.button(QDialogButtonBox.StandardButton.Cancel).setProperty("class", "secondary-button")
        button_box.button(QDialogButtonBox.StandardButton.Cancel).setText(self._("cancel"))
        return button_box

    def pass_values_through_validate(self,
                                     input_eigen_ref: str,
                                     input_bestek: str,
                                     input_subset: str,
                                     project: Project = None) -> None:
        """
        Validates input values and processes them for project upsert operations.

        This method checks the provided eigen reference, bestek, and subset values for validity
        using the HomeDomain validation methods. If the inputs are valid, it processes the upsert
        dialog input and closes the dialog; otherwise, it displays an error message.

        :param input_eigen_ref: The eigen reference input to validate.
        :type input_eigen_ref: str
        :param input_bestek: The bestek input to validate.
        :type input_bestek: str
        :param input_subset: The subset input to validate.
        :type input_subset: str
        :param project: An optional Project instance to process. Defaults to None.
        :type project: Project, optional

        :return: None
        """

        try:
            HomeDomain.validate(input_eigen_ref, input_bestek, input_subset)
        except EmptyFieldError as e:
            self.error_label.setText(str(e))
            return
        except WrongDatabaseError as e:
            self.error_label.setText(str(e))
            return
        except FileNotFoundError as e:
            self.error_label.setText(f"Can't find subset: {str(Path(input_subset).name)}")
            return
        self.error_label.setText("")

        try:
            HomeDomain.process_upsert_dialog_input(input_bestek, input_eigen_ref.strip(), input_subset, project)
            self.close()
        except ProjectExistsError as e:
            notification = NotificationWindow(title=self._("Project bestaat al"),
                                              message=self._(
                                                  "Project naam: \"{project_naam}\" bestaat al.\nGeef 1 van de projecten een andere naam of verwijder het bestaande project".format(
                                                      project_naam=e.eigen_referentie)))
            notification.exec()




    @staticmethod
    def open_file_picker(input_subset: QLineEdit):
        """
        Opens a file picker dialog to select a subset file.

        This static method displays a file dialog that allows the user to select a database file. If the input subset field is empty, the dialog starts in the user's home directory; otherwise, it starts in the directory specified by the input subset field.

        :param input_subset: The QLineEdit widget where the selected file path will be set.
        :type input_subset: QLineEdit

        :return: None
        """

        file_path = global_vars.get_start_dir_subset_selection(input_subset.text())

        selected_path_list = SubsetLoadFilePickerDialog.instance.summon(directory=file_path)
        if selected_path_list:
            selected = selected_path_list[0]
            input_subset.setText(str(selected))
            if selected:
                global_vars.last_subset_selected_dir = selected



import traceback
from pathlib import Path

from PyQt6.QtCore import QSize
from PyQt6.QtWidgets import QVBoxLayout, QFrame, QHBoxLayout, QLabel, QTreeWidget, QHeaderView, \
    QPushButton, QTableWidget, QTreeWidgetItem
from otlmow_converter.Exceptions.ExceptionsGroup import ExceptionsGroup

from otlmow_gui.Domain.logger.OTLLogger import OTLLogger
from otlmow_gui.Domain.step_domain.ExportFilteredDataSubDomain import ExportFilteredDataSubDomain
from otlmow_gui.GUI.dialog_windows.file_picker_dialog.LoadFilePickerDialog import LoadFilePickerDialog
from otlmow_gui.GUI.screens.ExportData_elements.ChangesTableView import ChangesTableView
from otlmow_gui.GUI.screens.ExportData_elements.TableErrorModel import TableErrorModel
from otlmow_gui.GUI.screens.ExportData_elements.TableModel import TableModel
from otlmow_gui.GUI.screens.ExportData_elements.AbstractExportDataSubScreen import AbstractExportDataSubScreen
from otlmow_gui.GUI.screens.general_elements.ButtonWidget import ButtonWidget
import qtawesome as qta

from otlmow_gui.GUI.translation.ValidationErrorReportTranslations import ValidationErrorReportTranslations
from otlmow_gui.exception_handler.ExceptionHandlers import create_task_reraise_exception


class ExportFilteredDataSubScreen(AbstractExportDataSubScreen):
    def __init__(self, language_settings=None):
        self.original_file_label = QLabel()
        self.original_file_field = QTreeWidget()

        self.control_button = ButtonWidget()
        self.refresh_button = QPushButton()
        # self.export_button = ButtonWidget()
        self.input_file_button = QPushButton()
        self.feedback_diff_label = QLabel()
        self.feedback_diff_table = ChangesTableView()
        self.message_icon = QLabel()
        self.message = QLabel()
        self.feedback_message_box = QFrame()
        self.model = None
        super().__init__(language_settings=language_settings)
        self.load_file_dialog_window = LoadFilePickerDialog(self._)

    def init_ui(self) -> None:
        """
        Initializes the user interface for the export data screen. This method sets up the layout
        by adding spacing, a menu, and configuring the container's margins and stretch properties
        to ensure proper alignment and appearance.

        :return: None
        """
        sub_window_layout = QVBoxLayout()
        sub_window_layout.addWidget(self.create_filter_unedited_data_box())
        sub_window_layout.addSpacing(10)
        sub_window_layout.addWidget(self.create_filetype_combobox())
        sub_window_layout.addWidget(self.create_relations_split_optionality())
        sub_window_layout.addSpacing(10)
        sub_window_layout.addWidget(self.create_button_box())
        sub_window_layout.addSpacing(10)

        self.setLayout(sub_window_layout)

    def create_filter_unedited_data_box(self):
        frame:QFrame = QFrame()
        layout:QHBoxLayout = QHBoxLayout()
        layout.addWidget(self.left_side(),stretch=1)
        layout.addWidget(self.right_side(),stretch=2)
        frame.setLayout(layout)
        return frame

    def left_side(self):
        frame = QFrame()
        frame_layout = QVBoxLayout()
        frame_layout.addWidget(self.input_original_file_field())
        frame_layout.addWidget(self.button_group())
        frame_layout.setContentsMargins(0, 0, 0, 0)
        frame.setLayout(frame_layout)
        return frame

    def right_side(self):
        frame = QFrame()
        frame_layout = QVBoxLayout()
        frame_layout.addWidget(self.create_feedback_diff_table())
        frame_layout.addWidget(self.feedback_message_box)
        frame_layout.setContentsMargins(0,0,0,0)
        frame.setLayout(frame_layout)
        return frame

    def create_feedback_diff_table(self):
        frame = QFrame()
        frame_layout = QVBoxLayout()
        self.feedback_diff_label.setText(self._('changes preview'))
        frame_layout.addWidget(self.feedback_diff_label)
        self.change_table()
        frame_layout.addWidget(self.feedback_diff_table)
        frame_layout.setContentsMargins(0, 0, 0, 0)
        self.construct_feedback_message()
        frame.setLayout(frame_layout)
        return frame

    def construct_feedback_message(self):
        # OTLLogger.logger.debug("constructing feedback message")
        frame_layout = QHBoxLayout()
        frame_layout.addWidget(self.message_icon)
        self.message.setProperty('class', 'feedback-message')
        frame_layout.addWidget(self.message)
        frame_layout.addStretch()
        frame_layout.setContentsMargins(0, 0, 0, 0)
        self.feedback_message_box.setLayout(frame_layout)

    def positive_feedback_message(self):
        self.message_icon.setVisible(True)
        self.message_icon.setPixmap(qta.icon('mdi.check', color="white").pixmap(QSize(48, 48)))
        self.message.setText(self._('all_info_correct'))
        self.feedback_message_box.setStyleSheet('background-color: #1DCA94; border-radius: 10px;')

    def clear_feedback_message(self):
        self.message.setText('')
        self.message_icon.setVisible(False)
        self.feedback_message_box.setStyleSheet('')

    def negative_feedback_message(self):
        self.message_icon.setVisible(True)
        self.message_icon.setPixmap(qta.icon('mdi.alert-circle-outline', color="white").pixmap(QSize(48, 48)))
        self.message.setText(self._('error'))
        self.feedback_message_box.setStyleSheet('background-color: #CC3300; border-radius: 10px;')

    def change_table(self):
        self.feedback_diff_table.setEnabled(True)
        self.feedback_diff_table.verticalHeader().setHidden(True)
        self.feedback_diff_table.horizontalHeader().setStretchLastSection(False)
        self.feedback_diff_table.resetPainting()
        self.feedback_diff_table.setShowGrid(False)
        self.feedback_diff_table.setEditTriggers(QTableWidget.EditTrigger.NoEditTriggers)


    def input_original_file_field(self):
        input_file = QFrame()
        input_file_layout = QVBoxLayout()
        self.original_file_label.setText(self._('original_file_load'))

        self.original_file_field.setHeaderHidden(True)
        self.original_file_field.setColumnCount(2)
        self.original_file_field.header().setStretchLastSection(False)
        self.original_file_field.header().setSectionResizeMode(0, QHeaderView.ResizeMode.Stretch)
        input_file_layout.addWidget(self.original_file_label)
        input_file_layout.addWidget(self.original_file_field)
        input_file_layout.setContentsMargins(0, 0, 0, 0)
        input_file.setLayout(input_file_layout)
        return input_file

    def button_group(self):
        frame = QFrame()
        frame_layout = QHBoxLayout()
        self.input_file_button.setText(self._('choose_file'))
        self.input_file_button.setProperty('class', 'primary-button')
        self.input_file_button.clicked.connect(lambda: self.open_original_file_picker())
        self.control_button.setText(self._('show differences'))
        self.control_button.setProperty('class', 'primary-button')
        self.control_button.setDisabled(True)
        self.control_button.clicked.connect(lambda: self.navigate_to_diff_report(self.feedback_diff_table))

        # self.export_button.setText(self._('apply differences'))
        # self.export_button.setDisabled(True)
        # self.export_button.setProperty('class', 'primary-button')
        # self.export_button.clicked.connect(lambda: self.replace_file_with_diff_report())

        self.refresh_button.setText(self._('empty fields'))
        self.refresh_button.setProperty('class', 'secondary-button')
        self.refresh_button.clicked.connect(lambda: self.remove_all_original_documents())
        frame_layout.addWidget(self.input_file_button)
        frame_layout.addStretch()
        frame_layout.addWidget(self.control_button)
        # frame_layout.addWidget(self.export_button)
        frame_layout.addWidget(self.refresh_button)
        frame_layout.setContentsMargins(0, 0, 0, 0)
        frame.setLayout(frame_layout)
        return frame

    def remove_all_original_documents(self):
        ExportFilteredDataSubDomain.remove_all_original_documents()


    def navigate_to_diff_report(self, table):
        try:
            create_task_reraise_exception(ExportFilteredDataSubDomain.get_diff_report())
        except Exception as e:
            # TODO: proper error message when original file to be loaded for comparison cannot be loaded
            raise e


    def fill_up_change_table(self, report):
        data = [
            [
                str(i+1),
                str(rep.id),
                str(rep.actie.value),
                str(rep.attribute),
                str(rep.original_value),
                str(rep.new_value),
            ]
            for i,rep in enumerate(report)
        ]
        self.model = TableModel(data, self._)
        self.feedback_diff_table.setModel(self.model)

    def clear_change_table(self):
        self.fill_up_change_table([])

    def open_original_file_picker(self):
        selected_file_path_list = self.load_file_dialog_window.summon()

        if selected_file_path_list:
            OTLLogger.logger.debug("file picker executed")
            ExportFilteredDataSubDomain.add_original_documents(selected_file_path_list)

            

    def update_original_files_list(self, files:dict[str,Path]):
        
        files_str = str(files)
        OTLLogger.logger.debug(f"adding file to list {files_str}")
        self.original_file_field.clear()
        self.clear_change_table()
        self.clear_feedback_message()

        for doc_name in files.keys():
            list_item = QTreeWidgetItem()
            list_item.setText(0, doc_name)
            list_item.setData(0, 1, doc_name)
            self.original_file_field.addTopLevelItem(list_item)
            button = ButtonWidget()
            button.clicked.connect(lambda: self.delete_file_from_list())
            button.setIcon(qta.icon('mdi.close'))
            self.original_file_field.setItemWidget(list_item, 1, button)



        if self.original_file_field.topLevelItemCount() == 0:
            self.control_button.setDisabled(True)
            self.export_btn.setDisabled(True)

        else:
            self.control_button.setDisabled(False)
            self.export_btn.setDisabled(False)

    def delete_file_from_list(self):
        items = self.original_file_field.selectedItems()
        doc_name = items[0].data(0,1)
        ExportFilteredDataSubDomain.delete_original_file(doc_name=doc_name)


    def process_export(self, document_path_list, csv_option, split_relations_and_objects):

        OTLLogger.logger.debug(document_path_list)
        create_task_reraise_exception(
            ExportFilteredDataSubDomain.export_diff_report(file_name=document_path_list[0],
                                                           separate_per_class_csv_option=csv_option,
                                                           separate_relations_option=split_relations_and_objects))

    def add_error_to_feedback_list(self, exception: Exception, doc: str) -> None:
        """Adds an error message to the feedback list based on the exception type.

        This method constructs a user-friendly error message from the provided
        exception and document name, then adds it to the feedback list. It
        handles various types of errors, formatting the message accordingly,
        and highlights the error in red for visibility.

        Args:
            self: The instance of the class.
            exception (Exception): The exception that occurred, used to determine the error message.
            doc (str): The path of the document associated with the error.

        Returns:
            None
        """

        OTLLogger.logger.debug(str(exception))
        traceback.print_exception(exception)
        doc_name = Path(doc).name

        return ValidationErrorReportTranslations.translate_exception(doc_name, exception)



    def fill_up_change_table_with_error_feedback(self, error_set: list[dict]):
        """Processes a set of errors and populates the feedback list.

        This method iterates through the provided error set, extracting exceptions
        and their associated document paths. It adds each error to the feedback
        list, handling both individual exceptions and groups of exceptions.

        Args:
            self: The instance of the class.
            error_set (list): A list of error items, each containing an exception
                              and a document path.

        Returns:
            None
        """
        data = []
        for item in error_set:
            exception = item["exception"]
            doc = item["path_str"]

            if isinstance(exception, ExceptionsGroup):
                for ex in exception.exceptions:
                    data.append([self.add_error_to_feedback_list(ex, doc)])
            else:
                data.append([self.add_error_to_feedback_list(exception, doc)])

        self.model = TableErrorModel(data, self._)
        self.feedback_diff_table.setModel(self.model)

    def create_button_box(self):
        frame = super().create_button_box()
        self.export_btn.setDisabled(True)
        return frame




import traceback
from pathlib import Path

from PyQt6.QtCore import Qt, QSize
from PyQt6.QtGui import QFont
from PyQt6.QtWidgets import QVBoxLayout, QWidget, QLabel, QPushButton, QFrame, QHBoxLayout, \
    QListWidget, \
    QListWidgetItem, QTreeWidget, QTreeWidgetItem, QHeaderView, QSizePolicy

from otlmow_gui.Domain.logger.OTLLogger import OTLLogger
from otlmow_converter.Exceptions.ExceptionsGroup import ExceptionsGroup
from otlmow_model.OtlmowModel.Helpers.OTLObjectHelper import count_assets_by_type

from otlmow_gui.Domain import global_vars
from otlmow_gui.Domain.enums import FileState
from otlmow_gui.Domain.step_domain.InsertDataDomain import InsertDataDomain
from otlmow_gui.GUI.dialog_windows.NotificationWindow import NotificationWindow
from otlmow_gui.GUI.dialog_windows.file_picker_dialog.LoadFilePickerDialog import LoadFilePickerDialog
from otlmow_gui.GUI.screens.general_elements.ButtonWidget import ButtonWidget
from otlmow_gui.GUI.dialog_windows.RemoveProjectFilesWindow import RemoveProjectFilesWindow
from otlmow_gui.GUI.dialog_windows.RevalidateDocumentsWindow import RevalidateDocumentsWindow
from otlmow_gui.GUI.screens.Screen import Screen
import qtawesome as qta

from otlmow_gui.GUI.translation.ValidationErrorReportTranslations import ValidationErrorReportTranslations
from otlmow_gui.exception_handler.ExceptionHandlers import create_task_reraise_exception


class InsertDataScreen(Screen):
    """
    Represents the screen for inserting and validating documents with objects.

    This class manages the user interface for inserting data, including 
    file selection, validation, and feedback display. It provides methods 
    for interacting with project files and displaying relevant messages to 
    the user.

    Args:
        language_settings (optional): Language settings for the user interface.

    Attributes:
        container_insert_data_screen (QVBoxLayout): Layout for the insert data screen.
        message_icon (QLabel): Label for displaying message icons.
        message (QLabel): Label for displaying messages to the user.
        input_file_label (QLabel): Label for the input file section.
        project_files_overview_field (QTreeWidget): Widget for displaying project files.
        feedback_message_box (QFrame): Frame for displaying feedback messages.
        asset_info (QListWidget): List widget for displaying asset information.
        input_file_button (ButtonWidget): Button for selecting input files.
        control_button (ButtonWidget): Button for controlling data processing.
        reset_button (ButtonWidget): Button for resetting the input fields.
        assets (list): List to store asset information.
        main_window (optional): Reference to the main application window.
    """
    feedback_message_icon_color = "white"

    def __init__(self, language_settings=None):
        super().__init__()
        self._ = language_settings

        self.load_file_dialog_window = LoadFilePickerDialog(self._)

        self.container_insert_data_screen = QVBoxLayout()

        self.message_icon = QLabel()
        self.message = QLabel()
        self.input_file_label = QLabel()

        self.project_files_overview_field: QTreeWidget = QTreeWidget()
        self.feedback_message_box = QFrame()
        self.asset_info = QListWidget()

        self.input_file_button = ButtonWidget()
        self.control_button = ButtonWidget()
        self.reset_button = ButtonWidget()

        self.assets = []
        self.main_window = None

        self.init_ui()


    def init_ui(self) -> None:
        """
        Sets up the user interface for the insert data screen.

        This method configures the layout by adding spacing and a menu to the
        insert data screen. It ensures that the UI elements are properly aligned
        and displayed within the designated container.

        :param self: The instance of the class.
        :returns: None
        """

        self.container_insert_data_screen.addSpacing(10)
        self.container_insert_data_screen.addWidget(self.create_menu())
        self.container_insert_data_screen.addSpacing(10)
        self.container_insert_data_screen.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.container_insert_data_screen)


    def create_menu(self) -> QWidget:  # sourcery skip: class-extract-method
        """
        Creates a menu layout for the insert data screen.

        This method constructs a QWidget that serves as a menu, organizing
        its components into a horizontal layout. It includes left and right
        sections that expand to fill available space, ensuring a responsive
        design.

        :param self: The instance of the class.
        :returns: QWidget -- The constructed menu widget.
        """

        window = QWidget()
        window.setProperty('class', 'background-box')
        window_layout = QHBoxLayout()

        left_side = self.left_side()
        left_side.setSizePolicy(QSizePolicy.Policy.Expanding,QSizePolicy.Policy.Expanding )

        right_side = self.right_side()
        right_side.setSizePolicy(QSizePolicy.Policy.Expanding,QSizePolicy.Policy.Expanding )

        window_layout.setContentsMargins(32, 0, 16, 0)
        window_layout.addWidget(left_side)
        window_layout.addWidget(right_side)

        window.setLayout(window_layout)

        return window


    def button_set(self) -> QFrame:
        """
        Creates and configures a frame containing control buttons.

        This method sets up a button frame with three buttons: a control button,
        a reset button, and a file selection button. Each button is configured
        with appropriate text, properties, and click event handlers to manage
        user interactions.

        :param self: The instance of the class.
        :returns: QFrame -- The configured button frame containing the buttons.
        """

        button_frame = QFrame()
        button_frame_layout = QHBoxLayout()

        self.control_button.setText(self._('control_button'))
        self.control_button.setDisabled(True)
        self.control_button.clicked.connect(lambda: self.try_to_validate_documents())
        self.control_button.setProperty('class', 'primary-button')

        reset_button = QPushButton()
        reset_button.setText(self._('reset_fields'))
        reset_button.setProperty('class', 'secondary-button')
        # noinspection PyUnresolvedReferences
        reset_button.clicked.connect(lambda: self.reset_button_functionality())

        self.input_file_button.setText(self._('choose_file'))
        self.input_file_button.setProperty('class', 'primary-button')
        self.input_file_button.clicked.connect(lambda: self.open_file_picker())

        button_frame_layout.addWidget(self.input_file_button)
        button_frame_layout.addStretch()
        button_frame_layout.addWidget(self.control_button)
        button_frame_layout.addWidget(reset_button)
        button_frame_layout.setContentsMargins(11, 11, 11, 0)

        button_frame.setLayout(button_frame_layout)

        self.warning_feedback_message()
        self.clear_feedback_message()

        return button_frame


    def try_to_validate_documents(self) -> None:
        """
        Attempts to validate documents and handle potential overwrites.

        This method checks if there is a recent quick save for the current project.
        If a quick save exists, it prompts the user with a warning about overwriting
        previous changes; otherwise, it proceeds to validate the documents.

        :param self: The instance of the class.
        :returns: None
        """

        # if there is a quick_save warns the user that they are overwriting the previous changes
        if global_vars.current_project.get_last_quick_save_path():
            RevalidateDocumentsWindow(self,self._)
        else:
            missing_project_files = InsertDataDomain.check_current_project_project_files_existence()
            if len(missing_project_files):
                self.show_missing_project_files_notification_window(missing_project_files)
                return

            create_task_reraise_exception(self.validate_documents())

    def show_missing_project_files_notification_window(self, missing_project_files):
        message = self._(
            "There are files missing.\nRemove these from the list and re-insert them:\n")
        for project_file in missing_project_files:
            filename = project_file.file_path.name
            message += f"   -{filename}\n"
        msgbox = NotificationWindow(message=message, title=self._("Missing project files"))
        msgbox.exec()

    async def validate_documents(self) -> None:
        """
        Validates documents and provides user feedback based on the results.

        This method clears any previous feedback and then attempts to load and
        validate the documents. Depending on whether errors are found, it
        provides appropriate feedback to the user and updates the UI accordingly.

        :param self: The instance of the class.
        :returns: None
        """

        self.clear_feedback()

        error_set, objects_list = await InsertDataDomain.load_and_validate_documents()
        self.asset_info.clear()
        if error_set:
            OTLLogger.logger.debug('negative feedback needed')
            self.negative_feedback_message()
            self.fill_error_feedback_list(error_set=error_set)
        else:
            OTLLogger.logger.debug('positive feedback needed')
            self.main_window.reset_ui(self._)
            self.positive_feedback_message()

        self.fill_feedback_list(objects_list)


    def fill_error_feedback_list(self, error_set: list[dict]):
        """Processes a set of errors and populates the feedback list.

        This method iterates through the provided error set, extracting exceptions
        and their associated document paths. It adds each error to the feedback
        list, handling both individual exceptions and groups of exceptions.

        Args:
            self: The instance of the class.
            error_set (list): A list of error items, each containing an exception
                              and a document path.

        Returns:
            None
        """



        for item in error_set:
            exception = item["exception"]
            doc = item["path_str"]

            if isinstance(exception, ExceptionsGroup):
                for ex in exception.exceptions:
                    self.add_error_to_feedback_list(ex, doc)
            else:
                self.add_error_to_feedback_list(exception, doc)


    def add_input_file_field(self) -> QFrame:
        """
        Creates and configures a frame for displaying project files.

        This method sets up a QFrame that contains a layout for displaying
        project files in a table format. It configures the table's column
        count, size policy, and header properties to ensure a user-friendly
        interface.

        Args:
            self: The instance of the class.

        Returns:
            QFrame: The configured frame containing the project files overview field.
        """

        input_file_frame = QFrame()
        input_file_layout = QHBoxLayout()

        self.project_files_overview_field.setColumnCount(3)
        self.project_files_overview_field.setSizePolicy(QSizePolicy.Policy.Expanding,QSizePolicy.Policy.Expanding )

        header = self.project_files_overview_field.header()
        header.setSectionResizeMode(1, QHeaderView.ResizeMode.Stretch)
        header.setStretchLastSection(False)
        self.project_files_overview_field.setHeaderHidden(True)

        input_file_layout.addWidget(self.project_files_overview_field)

        input_file_frame.setLayout(input_file_layout)

        return input_file_frame


    def left_side(self) -> QFrame:
        """
        Constructs the left side layout of the insert data screen.

        This method creates a QFrame that organizes various UI components
        vertically, including an input file label, input file field,
        control buttons, and a feedback message. It ensures proper spacing
        and layout for a cohesive user interface.

        Args:
            self: The instance of the class.

        Returns:
            QFrame: The configured frame representing the left side of the screen.
        """

        left_side = QFrame()
        left_side_layout = QVBoxLayout()
        left_side_layout.addSpacing(10)

        self.input_file_label.setText(self._('input_file'))

        left_side_layout.addWidget(self.input_file_label)
        left_side_layout.addWidget(self.add_input_file_field())
        left_side_layout.addWidget(self.button_set(),alignment=Qt.AlignmentFlag.AlignBottom)
        left_side_layout.addSpacing(10)
        left_side_layout.setStretch(2, 1)


        left_side.setLayout(left_side_layout)
        return left_side


    def right_side(self) -> QFrame:
        """
        Constructs the right side layout of the insert data screen.

        This method creates a QFrame that organizes various UI components
        vertically, including a list and a feedback message box. It ensures
        proper spacing and layout for a cohesive user interface.

        Args:
            self: The instance of the class.

        Returns:
            QFrame: The configured frame representing the right side of the screen.
        """

        right_side = QFrame()
        right_side_layout = QVBoxLayout()

        self.construct_feedback_message()

        right_side_layout.addSpacing(10)
        right_side_layout.addWidget(self.add_list())
        right_side_layout.addWidget(self.feedback_message_box)
        right_side_layout.addSpacing(10)
        right_side_layout.setStretch(1, 1)

        right_side.setLayout(right_side_layout)

        return right_side


    def construct_feedback_message(self) -> None:
        """
        Constructs and configures the feedback message display.

        This method sets up the layout for the feedback message box,
        including an icon and the message itself. It ensures that the
        message is styled appropriately for user visibility.

        Args:
            self: The instance of the class.

        Returns:
            None
        """

        OTLLogger.logger.debug("constructing feedback message")
        frame_layout = QHBoxLayout()

        self.message.setProperty('class', 'feedback-message')

        frame_layout.addWidget(self.message_icon)
        frame_layout.addWidget(self.message)

        self.feedback_message_box.setLayout(frame_layout)


    @classmethod
    def construct_dummy_feedback_message(cls) -> QFrame:
        """Creates a dummy feedback message display.

        This method constructs a QFrame that contains a placeholder for a
        feedback message. It sets up the layout to ensure that the dummy
        message is displayed correctly within the user interface.

        Args:
            cls: this class

        Returns:
            QFrame: The configured frame containing the dummy feedback message.
        """

        dummy_feedback = QFrame()
        frame_layout = QHBoxLayout()

        dummy_message = QLabel()
        dummy_message.setProperty('class', 'feedback-message')

        frame_layout.addWidget(QLabel())
        frame_layout.addWidget(dummy_message)

        dummy_feedback.setLayout(frame_layout)

        return dummy_feedback


    def add_list(self) -> QFrame:
        """Creates and configures a frame for displaying asset information.

        This method sets up a QFrame that contains a layout for displaying
        asset information. It ensures that the asset info widget is properly
        sized and positioned within the frame.

        Args:
            self: The instance of the class.

        Returns:
            QFrame: The configured frame containing the asset information.
        """

        frame = QFrame()
        frame_layout = QHBoxLayout()

        self.asset_info.setSizePolicy(QSizePolicy.Policy.Expanding,QSizePolicy.Policy.Expanding)

        frame_layout.addWidget(self.asset_info)
        frame_layout.setContentsMargins(0, 30, 0, 85)

        frame.setLayout(frame_layout)

        return frame


    def positive_feedback_message(self) -> None:
        """
        Displays a positive feedback message to the user.

        This method updates the feedback message box to indicate that all
        information is correct. It sets an appropriate icon, message text,
        and styles the message box for visual clarity.

        Args:
            self: The instance of the class.

        Returns:
            None
        """
        self.message_icon.setVisible(True)
        self.message_icon.setPixmap(qta.icon(
            'mdi.check',
            color=InsertDataScreen.feedback_message_icon_color).pixmap(QSize(48, 48)))
        self.message.setText(self._('all_info_correct'))
        self.feedback_message_box.setStyleSheet('background-color: #1DCA94; border-radius: 10px;')


    def warning_feedback_message(self) -> None:
        """
        Displays a warning feedback message to the user.

        This method updates the feedback message box to indicate a warning condition.
        It sets an appropriate icon, message text, and styles the message box to visually alert
        the user.

        Args:
            self: The instance of the class.

        Returns:
            None
        """
        self.message_icon.setVisible(True)
        self.message_icon.setPixmap(qta.icon(
            'mdi.alert',
            color=InsertDataScreen.feedback_message_icon_color).pixmap(QSize(48, 48)))
        self.message.setText(self._('warning'))
        self.feedback_message_box.setStyleSheet('background-color: #F8AA62; border-radius: 10px;')


    def clear_feedback_message(self) -> None:
        """
        Clears the feedback message display.

        This method resets the feedback message box by clearing the message
        text and removing any custom styles. It prepares the UI for new
        feedback messages.

        Args:
            self: The instance of the class.

        Returns:
            None
        """

        self.message.setText('')
        self.message_icon.setVisible(False)
        self.feedback_message_box.setStyleSheet('')



    def negative_feedback_message(self) -> None:
        """
        Displays a negative feedback message to the user.

        This method updates the feedback message box to indicate an error
        condition. It sets an appropriate icon, message text, and styles the
        message box to visually alert the user of the issue.

        Args:
            self: The instance of the class.

        Returns:
            None
        """
        self.message_icon.setVisible(True)
        self.message_icon.setPixmap(qta.icon(
            'mdi.alert-circle-outline',
            color=InsertDataScreen.feedback_message_icon_color).pixmap(QSize(48, 48)))
        self.message.setText(self._('error'))

        self.feedback_message_box.setStyleSheet('background-color: #CC3300; border-radius: 10px;')


    def reset_ui(self, language) -> None:
        super().reset_ui(self._)
        self._ = language
        self.input_file_label.setText(self._('input_file'))
        self.control_button.setText(self._('control_button'))
        self.clear_feedback()
        self.update_file_list()


    def open_file_picker(self) -> None:
        """
        Opens a file picker dialog for selecting files.

        This method initializes and displays a file picker dialog, allowing
        the user to select one or more files from their system. It sets the
        dialog's title, directory, and file filters based on supported file
        formats, and processes the selected files upon confirmation.

        Args:
            self: The instance of the class.

        Returns:
            None
        """

        selected_file_path_list = self.load_file_dialog_window.summon()

        if selected_file_path_list:
            try:
                InsertDataDomain.add_files_to_backend_list(files=selected_file_path_list)
            except Exception as e:
                # TODO: proper error message when file fails to be added to project
                raise e
            self.clear_feedback()


    def add_file_to_frontend_list(self,
                                  file: str,
                                  asset_state: FileState = FileState.WARNING) -> None:
        """Adds a file to the frontend list with a specified asset state.

        This method creates a new list item in the project files overview field
        for the specified file, setting its display name and icon based on the
        provided asset state. It also enables the control button and adds a
        delete button next to the list item for user interaction.

        Args:
            self: The instance of the class.
            file (str): The path of the file to be added to the list.
            asset_state (FileState, optional): The state of the asset, which
                                                determines the icon displayed.
                                                Defaults to FileState.WARNING.

        Returns:
            None
        """

        self.control_button.setDisabled(False)

        doc_name = Path(file).name

        list_item = QTreeWidgetItem()
        list_item.setText(1, doc_name)

        if asset_state == FileState.OK:
            list_item.setIcon(0, qta.icon('mdi.check', color="green"))
        elif asset_state == FileState.WARNING:
            list_item.setIcon(0, qta.icon('mdi.alert', color="orange"))
        elif asset_state == FileState.ERROR:
            list_item.setIcon(0, qta.icon('mdi.close', color="red"))

        list_item.setData(1, 1, file)
        list_item.setSizeHint(1, QSize(0, 30))

        button = ButtonWidget()
        button.clicked.connect(self.delete_file_from_list)
        button.setIcon(qta.icon('mdi.close'))

        self.project_files_overview_field.addTopLevelItem(list_item)
        self.project_files_overview_field.setItemWidget(list_item, 2, button)

    def add_file_overview_placeholder_to_front_end_list(self):
        file_place_holder_item = QTreeWidgetItem()
        file_place_holder_item.setText(1, self._("There are no files added to this project"))
        file_place_holder_item.setDisabled(True)


        placeholder_font = QFont()
        placeholder_font.setItalic(True)
        file_place_holder_item.setFont(1,placeholder_font)

        self.project_files_overview_field.addTopLevelItem(file_place_holder_item)




    def add_feedback_placeholder_to_front_end_list(self):
        place_holder_item = QListWidgetItem(
            self._(
                "Press validate to get feedback on the OTL-conformity of the files in this project"))

        placeholder_font = QFont()
        placeholder_font.setItalic(True)

        place_holder_item.setFont(placeholder_font)
        place_holder_item.setForeground(Qt.GlobalColor.gray)

        self.asset_info.addItem(place_holder_item)

    def delete_file_from_list(self) -> None:
        # sourcery skip: use-named-expression
        """
        Deletes the selected file from the project files list.

        This method retrieves the currently selected item in the project files
        overview field and deletes the corresponding file from the backend.
        If no item is selected, no action is taken.

        Args:
            self: The instance of the class.

        Returns:
            None
        """

        items = self.project_files_overview_field.selectedItems()

        if items:
            item_file_path = items[0].data(1,1)
            InsertDataDomain.delete_backend_document(item_file_path=item_file_path)

    def update_file_list(self) -> None:
        """
        Updates the frontend file list to synchronize with the backend.

        This method clears the current file list and synchronizes it with the
        backend documents. It also disables the control button if all documents
        are valid, indicating that no further actions can be taken.

        Args:
            self: The instance of the class.

        Returns:
            None
        """

        OTLLogger.logger.debug("[CLEAR] update_file_list")

        all_valid = InsertDataDomain.update_frontend()

    def update_control_button_state(self):
        self.control_button.setDisabled(self.project_files_overview_field.topLevelItemCount() == 0)

    def reset_button_functionality(self) -> None:
        """
        Resets the project by removing project files.

        This method opens a window to confirm the removal of project files associated with the
        current project and synchronizes the backend documents with the frontend.
        It also clears any existing feedback messages to provide a fresh state.

        Args:
            self: The instance of the class.

        Returns:
            None
        """

        RemoveProjectFilesWindow(project=global_vars.current_project, language_settings=self._)
        InsertDataDomain.update_frontend()
        self.clear_feedback()

    def clear_all(self) -> None:
        """
        Clears all feedback and resets the project files overview field.

        This method removes any feedback messages displayed to the user and
        clears the contents of the project files overview field, providing a
        fresh state for user interaction.

        Args:
            self: The instance of the class.

        Returns:
            None
        """

        self.clear_feedback()
        self.project_files_overview_field.clear()
        self.add_file_overview_placeholder_to_front_end_list()



    def clear_feedback(self) -> None:
        """
        Clears the feedback information displayed to the user.

        This method removes any asset information and resets the feedback
        message display, ensuring that the user interface is free of previous
        feedback.

        Args:
            self: The instance of the class.

        Returns:
            None
        """

        OTLLogger.logger.debug("[CLEAR] clear_feedback")
        self.asset_info.clear()
        self.add_feedback_placeholder_to_front_end_list()
        self.clear_feedback_message()

    def add_error_to_feedback_list(self, exception: Exception, doc: str) -> None:
        """Adds an error message to the feedback list based on the exception type.

        This method constructs a user-friendly error message from the provided
        exception and document name, then adds it to the feedback list. It
        handles various types of errors, formatting the message accordingly,
        and highlights the error in red for visibility.

        Args:
            self: The instance of the class.
            exception (Exception): The exception that occurred, used to determine the error message.
            doc (str): The path of the document associated with the error.

        Returns:
            None
        """

        OTLLogger.logger.debug(str(exception))
        traceback.print_exception(exception)
        doc_name = Path(doc).name
        error_widget = QListWidgetItem()

        error_text = ValidationErrorReportTranslations.translate_exception(doc_name, exception)

        error_widget.setText(error_text)
        self.asset_info.addItem(error_widget)

        item = self.asset_info.findItems(error_text, Qt.MatchFlag.MatchExactly)
        for item in item:
            self.asset_info.item(self.asset_info.row(item)).setForeground(Qt.GlobalColor.red)

    def fill_feedback_list(self, assets: list) -> None:
        """
        µFills the feedback list with asset information based on the provided assets.

        This method processes a list of assets, counting the number of objects
        by type and adding formatted messages to the feedback list. It also
        summarizes the total number of objects loaded that conform to the OTL
        standard.

        Args:
            self: The instance of the class.
            assets (list): A list of assets to be processed. If None, no action is taken.

        Returns:
            None
        """

        total_assets = 0
        if assets is None:
            return

        asset_dict = count_assets_by_type(objects=assets)
        for key, value in asset_dict.items():
            key_split = key.split('#')

            asset_widget = QListWidgetItem()
            asset_widget.setText(f'{value} objecten van het type {key_split[-1]} ingeladen\n')

            total_assets += value

            self.asset_info.addItem(asset_widget)

        asset_widget = QListWidgetItem()
        asset_widget.setText(
            f'In het totaal zijn er {total_assets} objecten ingeladen die conform zijn met de OTL '
            f'standaard\n')

        self.asset_info.addItem(asset_widget)



from otlmow_converter.Exceptions.UnknownExcelError import UnknownExcelError


class RelationHasInvalidTypeUriForSourceAndTarget(UnknownExcelError):
    """
    The bronAssetID.typeURI and doelAssetID.typeURI are INVALID for the current relationTypeURI
    For this error to occur the bronAssetID.typeURI or doelAssetID.typeURI need to exist first otherwise
    Error RelationHasNonExistingTypeUriForSourceOrTarget is thrown
    """
    def __init__(self,relation_type_uri,relation_identificator,wrong_field,wrong_value,wrong_field2= None,wrong_value2 = None,tab=None,*args):
        super().__init__(Exception(),*args)
        self.relation_type_uri = relation_type_uri
        self.relation_identificator = relation_identificator
        self.wrong_field = wrong_field
        self.wrong_value = wrong_value
        self.wrong_field2 = wrong_field2
        self.wrong_value2 = wrong_value2
        self.tab = tab




import os
import shutil
from pathlib import Path
from typing import List

import requests
from github import Github

from otlmow_gui.Domain.logger.OTLLogger import OTLLogger


class GitHubDownloader:
    def __init__(self, repo_owner_and_name: str):
        self.repo_owner_and_name = repo_owner_and_name
        self.repo: Github | None = None

    def get_repo(self):
        """
        Retrieves the GitHub repository associated with the specified owner and name.
        If the repository has not been previously fetched, it will be retrieved from the GitHub API.

        Returns:
            Repository: The GitHub repository object.

        Raises:
            Exception: If there is an issue connecting to the GitHub API or retrieving the repository.
        """

        if self.repo is None:
            self.repo = Github().get_repo(self.repo_owner_and_name)
        return self.repo

    def download_full_repo(self, dest: Path):
        """
        Downloads a zipped version of the entire GitHub repository to a specified destination directory.
        It ensures that the destination directory exists before saving the downloaded file.

        Args:
            dest (Path): The directory where the zipped repository will be saved.

        Returns:
            None

        Raises:
            OSError: If the destination directory cannot be created or accessed.
        """
        dest_filename = 'full_repo_download.zip'
        dest_dir = dest.parent / dest.stem
        #dest can be a directory or a zip file
        if dest.suffix == ".zip":
            dest_dir = dest.parent
            dest_filename = dest.name

        if  os.path.exists(dest_dir):
            shutil.rmtree(dest_dir)
        os.makedirs(dest_dir)

        dest = dest_dir / dest_filename
        OTLLogger.logger.info(f"Download otlmow-model to zip: {dest}",extra={"timing_ref":"download_otlmow_model"})

        download_zip_url = f'https://github.com/{self.repo_owner_and_name}/zipball/master'
        resp = requests.get(download_zip_url)

        with open(dest, 'wb') as f:
            f.write(resp.content)
        OTLLogger.logger.info("Download otlmow-model to zip",
                              extra={"timing_ref": "download_otlmow_model"})

    def download_file_to_dir(self, file_path: str, destination_dir: Path, contents=None):
        """
        Downloads a file from a GitHub repository and saves it to a specified directory.
        If the file contents are not provided, it retrieves them from the repository.

        Args:
            destination_dir (Path): The directory where the downloaded file will be saved.
            file_path (str, optional): The path of the file in the repository. Defaults to None.
            contents (optional): The contents of the file to be downloaded. If None, it will be fetched from the repository.

        Returns:
            None

        Raises:
            OSError: If the destination directory cannot be created or accessed.
        """

        if contents is None:
            contents = self.get_repo().get_contents(file_path)

        if not os.path.exists(destination_dir):
            os.makedirs(destination_dir)

        url = f'https://raw.githubusercontent.com/{self.repo_owner_and_name}/master/{file_path}'
        resp = requests.get(url)
        with open(destination_dir / contents.name, 'wb') as f:
            f.write(resp.content)

    def download(self, dir_or_file_path: str, destination_dir: Path):
        """
        Downloads a file or directory from a GitHub repository to a specified destination directory.
        It checks the type of the contents and delegates the download process accordingly.

        Args:
            dir_or_file_path (str): The path of the file or directory in the repository to be downloaded.
            destination_dir (Path): The directory where the downloaded content will be saved.

        Returns:
            None

        Raises:
            OSError: If the destination directory cannot be created or accessed.
        """

        if not os.path.exists(destination_dir):
            os.makedirs(destination_dir)

        contents = self.get_repo().get_contents(dir_or_file_path)

        if isinstance(contents, list) or contents.type == 'dir':
            self.download_dir(None, destination_dir=destination_dir, contents=contents)
        else:
            self.download_file_to_dir(None, destination_dir=destination_dir, contents=contents)

    def download_dir(self, dir_path: str, destination_dir: Path, contents: List = None):
        """
        Downloads all files and subdirectories from a specified directory in a GitHub repository to a local destination.
        It recursively processes each item, ensuring that directories are handled appropriately.

        Args:
            destination_dir (Path): The directory where the downloaded contents will be saved.
            dir_path (str, optional): The path of the directory in the repository to be downloaded. Defaults to None.
            contents (List, optional): A list of contents to download. If None, it will be fetched from the repository.

        Returns:
            None

        Raises:
            OSError: If the destination directory cannot be created or accessed.
        """
        if contents is None:
            contents = self.get_repo().get_contents(dir_path)
        for content in contents:
            if content.type == 'dir':
                self.download_dir(dir_path=content.path, destination_dir=destination_dir)
            else:
                self.download_file_to_dir(file_path=content.path, destination_dir=destination_dir)

    def download_file_to_memory(self, file_path: str = None):
        """
        Downloads a file from a GitHub repository and returns its contents in memory.
        It raises an error if the file cannot be accessed, ensuring that only successful downloads are processed.

        Args:
            file_path (str, optional): The path of the file in the repository to be downloaded. Defaults to None.

        Returns:
            bytes: The contents of the downloaded file.

        Raises:
            ConnectionRefusedError: If the file cannot be accessed due to a connection issue.
        """

        url = f'https://raw.githubusercontent.com/{self.repo_owner_and_name}/master/{file_path}'
        OTLLogger.logger.info(f"Downloading file from {url}")
        resp = requests.get(url, timeout=5)
        if resp.status_code != 200:
            raise ConnectionRefusedError()

        return resp.content


import asyncio
import importlib
import typing

from datetime import datetime
from pathlib import Path

import sys
import os

from otlmow_gui.Domain.ProgramFileStructure import ProgramFileStructure

if sys.platform.startswith("linux"):
    os.environ["QT_QUICK_BACKEND"] = "software" # fixes a flickering issue on some linux systems

from PyQt6 import QtCore

from otlmow_gui.Domain import global_vars
from otlmow_gui.Domain.Settings import Settings
from otlmow_gui.Domain.logger.OTLLogger import OTLLogger
from otlmow_gui.Domain.step_domain import InsertDataDomain
from otlmow_gui.Domain.network.Updater import Updater
from otlmow_gui.GUI.Styling import Styling
from otlmow_gui.GUI.translation.GlobalTranslate import GlobalTranslate
from otlmow_gui.exception_handler.ExceptionHandlers import create_task_reraise_exception, excepthook

ROOT_DIR =  Path(Path(__file__).absolute()).parent
sys.path.insert(0,str(ROOT_DIR.absolute()))# needed for python to import project files



from PyQt6.QtGui import QIcon
from PyQt6.QtWidgets import QApplication
from qasync import QEventLoop, asyncClose




from otlmow_gui.Domain.project import Project
from otlmow_gui.GUI.MainWindow import MainWindow

project_dir = ROOT_DIR / 'demo_projects/'

# Decide where the locale folder lives
if getattr(sys, 'frozen', False):
    # Running from PyInstaller bundle
    base_path = Path(sys._MEIPASS)  # temp dir PyInstaller unpacks to
else:
    # Running from source
    base_path = ROOT_DIR

LANG_DIR = base_path / 'locale'


    # OTLLogger.logger.info('Splash screen closed.')

# Used to add demo data to the application for showcase purpose only
def demo_data():
    project_1 = Project(eigen_referentie="test1", project_path=Path(
        Path.home() / 'OTLWizardProjects' / 'Projects' / 'project_1'),
                        subset_path=Path(project_dir / 'project_1' / 'Flitspaal_noAgent3.0.db'),
                        saved_documents_overview_path=Path(project_dir / 'project_1'),
                        bestek="test_bestek1", laatst_bewerkt=datetime(2021, 9, 11))
    project_1.save_project_to_dir()
    return project_1


class OTLWizard(QApplication):

    def __init__(self,settings: dict, argv: typing.List[str]):
        super().__init__(argv)
        # Windows will set the colorScheme to Dark if that is the setting of the system
        # We will set it to light no matter what the colorscheme of the system is.
        # if self.styleHints().colorScheme() == Qt.ColorScheme.Dark:
        # self.styleHints().setColorScheme(Qt.ColorScheme.Light)

        sys.excepthook = excepthook


        Updater.update_oltmow_model()

        if '_PYI_SPLASH_IPC' in os.environ and importlib.util.find_spec("pyi_splash"):
            import pyi_splash
            pyi_splash.update_text('UI Loaded ...')
            pyi_splash.close()

        Updater.check_for_OTL_wizard_updates()

        IMG_DIR = ProgramFileStructure.get_dynamic_library_path('img')

        app_icon = QIcon(str(IMG_DIR / 'wizard.ico'))
        self.setWindowIcon(app_icon)

        language = GlobalTranslate(settings, LANG_DIR).get_all()
        global_vars.otl_wizard = self
        self.main_window = MainWindow(language)
        self.main_window.resize(1250, 650)
        self.main_window.setWindowTitle('OTLWizard')
        self.main_window.setMinimumSize(800, 600)
        self.main_window.setWindowFlags(
            self.main_window.windowFlags() | QtCore.Qt.WindowStaysOnTopHint)  # set always on top flag, makes window disappear
        self.main_window.show()



        self.meipass = sys._MEIPASS if hasattr(sys, '_MEIPASS') else None
        Styling.applyStyling(self,self.meipass)
        self.paletteChanged.connect(lambda state: Styling.applyStyling(self,self.meipass))


        # self.demo_project = demo_data()
        self.demo_project = None

        if "--test" in argv:
            self.test_setup()

    def test_setup(self):
        self.main_window.home_screen.table.open_project(2)
        create_task_reraise_exception(InsertDataDomain.load_and_validate_documents())
        self.main_window.setCurrentIndex(3)
        self.main_window.reset_ui(self.main_window._)
        self.main_window.step_3_tabwidget.tabs.setCurrentWidget(self.main_window.step3_relations)

    @asyncClose
    async def quit(self):
        OTLLogger.logger.debug("closing application")
        if self.demo_project:
            self.demo_project.delete_project_dir_by_path()
        super().quit()


def otl_wizard_2_main():
    program_settings = Settings.get_or_create_settings_file()
    OTLLogger.init()
    OTLLogger.logger.debug("Application started")

    app = OTLWizard(program_settings, sys.argv)

    if '_PYI_SPLASH_IPC' in os.environ and importlib.util.find_spec("pyi_splash"):
        import pyi_splash

        pyi_splash.update_text('UI Loaded ...')
        pyi_splash.close()
        OTLLogger.logger.info('Splash screen closed.')

    app.main_window.raise_()
    app.main_window.setWindowFlags(
        app.main_window.windowFlags() & ~QtCore.Qt.WindowStaysOnTopHint)  # clear always on top flag, makes window disappear
    app.main_window.show()
    event_loop = QEventLoop(app)
    asyncio.set_event_loop(event_loop)

    event_loop.run_forever()


import logging
import webbrowser

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QDialog, QVBoxLayout, QLabel, QDialogButtonBox


class SuggestUpdateWindow:


    def __init__(self, language_settings, local_version:str, new_version:str,otl_model_out_of_date:bool = False):
        self.local_version:str = local_version
        self.new_version:str = new_version
        self._ = language_settings

        if otl_model_out_of_date:
            self.dialog_title:str = self._("otl_model_out_of_date_title")
            self.dialog_text:str = self._("otl_model_out_of_date_text").format(self.local_version,self.new_version)
        else:
            self.dialog_title:str  = self._("update_available_title")
            self.dialog_text:str  = self._("update_available_text").format(self.local_version,self.new_version)
        self.init_ui()

    def init_ui(self):
        dialog = QDialog()
        dialog.setModal(True)
        dialog.setWindowTitle(self.dialog_title)
        layout = QVBoxLayout()
        question_label = QLabel()
        question_label.setText(self.dialog_text)
        layout.addWidget(question_label)
        button_box = self.create_button_box()
        button_box.accepted.connect(lambda: self.remove_project_files(dialog))
        button_box.rejected.connect(dialog.reject)
        layout.addWidget(button_box)
        dialog.setLayout(layout)

        dialog.setWindowFlags(dialog.windowFlags() | Qt.WindowStaysOnTopHint)  # Keep it on top

        dialog.raise_()
        dialog.activateWindow()

        dialog.exec()

    def create_button_box(self):
        button_box = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)
        button_box.setProperty("class", "button-box")
        button_box.button(QDialogButtonBox.StandardButton.Ok).setProperty("class", "primary-button")
        button_box.button(QDialogButtonBox.StandardButton.Ok).setText(self._("yes"))
        button_box.button(QDialogButtonBox.StandardButton.Cancel).setProperty("class", "secondary-button")
        button_box.button(QDialogButtonBox.StandardButton.Cancel).setText(self._("no"))
        return button_box

    def remove_project_files(self, dialog):
        logging.info("User choose to update")
        self.open_wiki()
        dialog.close()


    @staticmethod
    def open_wiki():
        webbrowser.open('https://github.com/davidvlaminck/otlmow-gui/wiki/Installation')



from pathlib import Path
from typing import Optional

from PyQt6.QtWidgets import QFileDialog

from otlmow_gui.Domain import global_vars
from otlmow_gui.Domain.logger.OTLLogger import OTLLogger
from otlmow_gui.GUI.dialog_windows.file_picker_dialog.AbstractFilePickerDialog import AbstractFilePickerDialog


class LoadFilePickerDialog(AbstractFilePickerDialog):
    def __init__(self, language_settings,title:Optional[str]= None,name_filter:str=None, initial_file_path: Optional[str]=None):



        super().__init__(language_settings, title, name_filter, initial_file_path)
        self.setFileMode(QFileDialog.FileMode.ExistingFiles)

        if not title:
            self.setWindowTitle(self._("Selecteer bestand"))  # standard title for load dialogs

    def getCustomFilters(self) -> str:
        supported_extensions = [f'*.{value}' for value in
                                global_vars.supported_file_formats.values()]
        supported_extensions_string = ' '.join(supported_extensions)
        filters = [self._('Alle ondersteunde bestanden ({supported_extensions_string})').format(
            supported_extensions_string=supported_extensions_string)]

        filters.extend(f"{k} files (*.{v})" for k, v in global_vars.supported_file_formats.items())

        return ";;".join(filters)

    def execute(self) -> list[Path]:
        if self.exec():
            res = [Path(selected_path_str) for selected_path_str in self.selectedFiles() if
             selected_path_str]
            OTLLogger.logger.debug(f"save_file_dialog return: {res}")
            return res
        OTLLogger.logger.debug(f"save_file_dialog return: {[]}")
        return []




from typing import Any

from PyQt6.QtCore import QAbstractTableModel, Qt, QModelIndex


class TableErrorModel(QAbstractTableModel):
    def __init__(self, data, language_settings):
        super(TableErrorModel, self).__init__()
        self._data = data
        self._ = language_settings
        self.header_labels = [self._('error')]

    def data(self, index: QModelIndex, role: int = ...) -> Any:
        if role == Qt.ItemDataRole.DisplayRole:
            return self._data[index.row()][index.column()]

    def rowCount(self, parent: QModelIndex = ...) -> int:
        return len(self._data)

    def columnCount(self, parent: QModelIndex = ...) -> int:
        if self._data and self._data[0]:
            return len(self._data[0])
        else:
            return 0

    def headerData(self, section: int, orientation: Qt.Orientation, role: int = ...) -> Any:
        if role == Qt.ItemDataRole.DisplayRole and orientation == Qt.Orientation.Horizontal:
            return self.header_labels[section]
        return QAbstractTableModel.headerData(self, section, orientation, role)



from PyQt6.QtWidgets import QVBoxLayout, QFrame, QLabel, QHBoxLayout, QLineEdit, QPushButton


from otlmow_gui.GUI.screens.Screen import Screen
import qtawesome as qta


class ConversionScreen(Screen):
    def __init__(self, language_settings=None):
        super().__init__()
        self._ = language_settings
        self.container_insert_data_screen = QVBoxLayout()
        self.input_file_label = QLabel()
        self.convert_btn = QPushButton()
        self.init_ui()

    def init_ui(self):
        self.container_insert_data_screen.addSpacing(10)
        self.container_insert_data_screen.addWidget(self.create_menu())
        self.container_insert_data_screen.addStretch()
        self.container_insert_data_screen.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.container_insert_data_screen)

    def create_menu(self):
        window = QFrame()
        window_layout = QVBoxLayout()
        window.setProperty('class', 'background-box')
        title = QLabel()
        title.setText(self._('file_conversion'))
        title.setProperty('class', 'sub-title')

        window_layout.addWidget(title)
        window_layout.addSpacing(20)
        window_layout.addWidget(self.input_file_field())
        window_layout.addSpacing(10)
        window_layout.addWidget(self.button_box())
        window_layout.addSpacing(10)
        window.setLayout(window_layout)
        return window

    def button_box(self):
        button_box = QFrame()
        button_box_layout = QHBoxLayout()
        self.convert_btn.setText(self._('convert'))
        self.convert_btn.setProperty('class', 'primary-button')
        button_box_layout.addWidget(self.convert_btn)
        button_box_layout.addStretch()
        button_box.setLayout(button_box_layout)
        return button_box

    def input_file_field(self):
        input_file = QFrame()
        input_file_layout = QHBoxLayout()
        self.input_file_label.setText(self._('file_for_conversion'))
        input_file_field = QLineEdit()
        input_file_field.setReadOnly(True)
        input_file_button = QPushButton()
        input_file_button.setIcon(qta.icon('mdi.folder-open-outline'))
        input_file_layout.addWidget(self.input_file_label)
        input_file_layout.addWidget(input_file_field)
        input_file_layout.addWidget(input_file_button)
        input_file.setLayout(input_file_layout)
        return input_file

    def reset_ui(self, _):
        super().reset_ui(_)
        self._ = _
        self.input_file_label.setText(self._('file_for_conversion'))
        self.convert_btn.setText(self._('convert'))



import sqlite3
from pathlib import Path
from sqlite3 import Connection

from otlmow_gui.Domain.logger.OTLLogger import OTLLogger
from otlmow_gui.Exceptions.NotASqlliteFileError import NotASqlliteFileError


class SubsetDatabase:

    def __init__(self, db_path: Path):
        self.db_path: Path = db_path
        self.connection: Connection
        self.create_connection(self.db_path)

    def create_connection(self, db_path: Path) -> None:
        if db_path.exists():
            try:
                self.connection = sqlite3.connect(db_path)
            except sqlite3.OperationalError as e:
                OTLLogger.logger.error(e)
                raise NotASqlliteFileError(e)
        else:
            raise FileNotFoundError(f'{db_path} is not a valid path. File does not exist.')

    def get_general_info_project(self) -> list:
        cursor = self.connection.cursor()
        cursor.execute("SELECT * FROM GeneralInfo")
        info = cursor.fetchall()
        cursor.close()
        return info

    def is_valid_subset_database(self) -> bool:
        try:
            cursor = self.connection.cursor()
            cursor.execute("SELECT * FROM OSLOClass")
            cursor.fetchall()
            cursor.close()
            return True
        except sqlite3.OperationalError:
            return False

    def close_connection(self) -> None:
        OTLLogger.logger.debug("Closing connection to subset database")
        self.connection.close()



import asyncio
import os
import subprocess
import webbrowser
from asyncio import sleep
from datetime import datetime
from pathlib import Path

import qtawesome as qta
from PyQt6.QtCore import Qt
from PyQt6.QtGui import QPixmap, QTransform
from PyQt6.QtWidgets import QFrame, QHBoxLayout, QLabel, QVBoxLayout, QWidget, QMenu

from otlmow_gui.Domain import global_vars
from otlmow_gui.Domain.ProgramFileStructure import ProgramFileStructure
from otlmow_gui.Domain.logger.OTLLogger import OTLLogger
from otlmow_gui.Domain.project.Project import Project
from otlmow_gui.Domain.step_domain.HomeDomain import HomeDomain
from otlmow_gui.GUI.Styling import Styling
from otlmow_gui.GUI.dialog_windows.NotificationWindow import NotificationWindow
from otlmow_gui.GUI.dialog_windows.ProjectExistsError import ProjectExistsError
from otlmow_gui.GUI.dialog_windows.file_picker_dialog.ProjectImportFilePickerDialog import \
    ProjectImportFilePickerDialog
from otlmow_gui.GUI.screens.general_elements.ButtonWidget import ButtonWidget
from otlmow_gui.GUI.dialog_windows.LanguageWindow import LanguageWindow
from otlmow_gui.GUI.dialog_windows.MenuActionsWindow import MenuActionsWindow

from otlmow_gui.GUI.dialog_windows.UpsertProjectWindow import UpsertProjectWindow
from otlmow_gui.GUI.screens.Home_elements.OverviewTable import OverviewTable


IMG_DIR = ProgramFileStructure.get_dynamic_library_path('img')

class HeaderBar(QFrame):

    def __init__(self, language, main_window=None, table: OverviewTable = None, has_save_btn: bool = True):
        super().__init__()
        self.new_project_button = ButtonWidget()
        self._ = language
        self.table = table
        self.main_window = main_window
        self.return_button = ButtonWidget()
        self.subtitle = QLabel()
        self.save_button = ButtonWidget()
        self.import_button = ButtonWidget()
        self.reference_title = QLabel()
        self.has_save_btn = has_save_btn
        self.loading_icon = QLabel()
        self.saving_msg = QLabel()
        self.animation_event_loop = None
        self.pixmap_icon = None
        self.initialised = False
        self.quick_save_animation_task = None
        self.header = None
        self.import_project_dialog_window: ProjectImportFilePickerDialog = (
            ProjectImportFilePickerDialog(self._))

    def construct_header_bar(self):
        self.setProperty('class', 'header')
        self.header = QHBoxLayout()
        title = QLabel('OTLWizard')
        title.setProperty('class', 'title')
        self.header.addWidget(title)
        self.header.addSpacing(30)
        self.create_new_project_button()
        self.create_import_button()
        self.header.addWidget(self.new_project_button)
        self.header.addWidget(self.import_button)
        # self.header.addWidget(self.create_loading_icon())
        self.header.addStretch()
        self.header.setAlignment(self.new_project_button, Qt.AlignmentFlag.AlignLeft)
        user_settings = self.construct_settings_bar()
        self.header.addLayout(user_settings)
        self.header.setAlignment(user_settings, Qt.AlignmentFlag.AlignRight)
        self.setLayout(self.header)

    def start_event_loop(self):

        self.animation_event_loop = asyncio.get_event_loop()
        if self.quick_save_animation_task and self.animation_event_loop.is_running():
            self.quick_save_animation_task.cancel()

        self.quick_save_animation_task = self.animation_event_loop.create_task(self.rotate_icon())

    def create_new_project_button(self):
        self.new_project_button.setIcon(qta.icon("mdi.plus",
                                                 color='white'))
        self.new_project_button.setText(self._('new_project_button'))
        self.new_project_button.setProperty('class', 'new-project-btn')
        self.new_project_button.clicked.connect(
            lambda: self.start_dialog_window(is_project=True))

    def create_loading_icon(self):
        self.pixmap_icon = QPixmap(str(IMG_DIR / 'wizard.ico'))

        header_sub = QFrame()
        header_sub_layout = QHBoxLayout()

        self.loading_icon.setPixmap(self.pixmap_icon )
        self.loading_icon.setFixedHeight(30)
        self.loading_icon.setFixedWidth(30)
        self.loading_icon.setScaledContents(True)
        self.saving_msg.setText("Unsaved")
        self.saving_msg.setProperty("class","white_text")
        header_sub_layout.addWidget(self.loading_icon)
        header_sub_layout.addWidget(self.saving_msg)
        header_sub_layout.setContentsMargins(0, 0, 0, 0)
        header_sub.setLayout(header_sub_layout)

        return header_sub

    async def rotate_icon(self):
        angle = 0
        count = 0
        time_string = datetime.now().strftime("%H:%M")
        self.saving_msg.setText("Saving")
        while count < (360/10): # about 1.8 seconds
            transform = QTransform()
            angle += 10
            count +=1
            angle = angle % 360
            transform.rotate(angle)
            rotation_pixmap = self.pixmap_icon.transformed(transform)
            self.loading_icon.setPixmap(rotation_pixmap)
            if(count == 20):
                self.saving_msg.setText(f"Last Saved at {time_string}")
            # elif(count%5 == 3):
            #     self.saving_msg.setProperty("class", "black_text")

            await sleep(0.05)

        self.loading_icon.setPixmap(self.pixmap_icon)
        # self.saving_msg.setText(f"Last Saved at {time_string}")

    def construct_settings_bar(self):
        user_pref_container = QHBoxLayout()
        settings = ButtonWidget()
        settings.setIcon(qta.icon('mdi.cog',
                                  color="white"))
        settings.setProperty('class', 'settings')
        settings.clicked.connect(lambda: self.start_dialog_window())
        user_pref_container.addWidget(settings)
        help_widget = ButtonWidget()
        menu = self.construct_menu()
        help_widget.setMenu(menu)
        help_icon = qta.icon('mdi.help-circle',
                             color='white')
        help_widget.setIcon(help_icon)
        help_widget.setProperty('class', 'settings')
        user_pref_container.addWidget(help_widget)
        return user_pref_container

    def construct_menu(self):
        menu = QMenu()
        self.help_action = menu.addAction(self._('help'))
        self.help_action.triggered.connect(lambda: self.open_wiki())
        self.about_action = menu.addAction(self._('about'))
        self.about_action.triggered.connect(lambda: MenuActionsWindow(self._).create_about_window())
        self.report_action = menu.addAction(self._('report error'))
        self.report_action.triggered.connect(lambda: MenuActionsWindow(self._).create_error_report_window())
        self.logs_action = menu.addAction(self._('logs'))
        self.logs_action.triggered.connect(self.open_logs_folder)

        return menu

    @classmethod
    def open_logs_folder(cls):
        logs_folderpath = Path.home() / 'OTLWizardProjects' / 'logs'
        log_filename_list = os.listdir(logs_folderpath)
        if log_filename_list:
            log_filename_list.sort(reverse=True)
            last_log_filepath = logs_folderpath / log_filename_list[0]
            command = f'explorer /select,"{last_log_filepath}"'
            OTLLogger.logger.debug(f"open logs : {command}")
            subprocess.Popen(command)
        else:
            command = f'explorer /select,"{logs_folderpath}"'
            OTLLogger.logger.debug(f"open logs : {command}")
            subprocess.Popen(command)

    @staticmethod
    def open_wiki():
        webbrowser.open('https://github.com/davidvlaminck/OTLMOW-GUI/wiki')

    def start_dialog_window(self, id_: int = None, is_project=False) -> None:
        if is_project:
            dialog = UpsertProjectWindow(language_settings=self._,project=id_)
        else:
            dialog = LanguageWindow(language_settings=self._,main_window=self.main_window)
        dialog.exec()

    def header_bar_detail_screen(self):
        if self.initialised:
            return self.return_button

        self.return_button = ButtonWidget()
        full_header = QVBoxLayout()
        header = QHBoxLayout()
        head_top = QWidget()
        head_top.setProperty('class', 'header')
        title = QLabel('OTLWizard')
        title.setProperty('class', 'title')
        header.addWidget(title)
        header.addSpacing(20)
        self.return_button.setProperty('class', 'return-button')
        self.return_button.setIcon(qta.icon('mdi.arrow-left',
                                            color='#B35F35'))
        self.return_button.setText(self._('return_to_home_screen'))
        if global_vars.current_project is not None:
            self.reference_title.setText(global_vars.current_project.eigen_referentie)
        else:
            self.reference_title.setText("")
        self.reference_title.setProperty('class', 'project-title')
        header.addWidget(self.reference_title)
        header.addWidget(self.create_loading_icon())
        header.addStretch()
        header.setAlignment(self.return_button, Qt.AlignmentFlag.AlignLeft)
        settings = self.construct_settings_bar()
        header.addLayout(settings)
        header.setAlignment(settings, Qt.AlignmentFlag.AlignRight)

        header_sub = QFrame()
        header_sub_layout = QHBoxLayout()
        header_sub.setProperty('class', 'sub-header')
        if self.has_save_btn:
            self.save_button.setIcon(qta.icon('mdi.content-save',
                                              color='white'))
            self.save_button.setText(self._('save_button'))
            self.save_button.setProperty('class', 'primary-button')

        header_sub_layout.addWidget(self.return_button, alignment=Qt.AlignmentFlag.AlignLeft)
        if self.has_save_btn:
            header_sub_layout.addWidget(self.save_button)
            header_sub_layout.setAlignment(self.save_button, Qt.AlignmentFlag.AlignRight)

        header_sub.setLayout(header_sub_layout)

        head_top.setLayout(header)

        full_header.setSpacing(0)
        full_header.addWidget(head_top)
        full_header.addWidget(header_sub)
        full_header.setContentsMargins(0, 0, 0, 0)
        self.setLayout(full_header)
        self.initialised = True
        return self.return_button



    def create_import_button(self):
        self.set_import_icon(self.import_button)
        self.import_button.setText(self._("import"))
        self.import_button.setProperty('class', 'import-button')
        self.import_button.clicked.connect(
            lambda: self.import_project_window())

    def import_project_window(self):
        selected_file_path_list  = self.import_project_dialog_window.summon()

        if not selected_file_path_list or not selected_file_path_list[0]:
            return

        project = None
        try:
            project = Project.import_project(selected_file_path_list[0])
        except ProjectExistsError as e:
            notification = NotificationWindow(title=self._("Project bestaat al"),
                                              message=self._("Project naam: \"{project_naam}\" bestaat al.\nVerwijder het bestaande project en importeer opnieuw".format(project_naam=e.eigen_referentie)))
            notification.exec()
        except Exception as e:
            # TODO: proper error messag when project fails to import
            raise e


        if project is None:
            return
        HomeDomain.projects[project.eigen_referentie] = project
        global_vars.otl_wizard.main_window.home_screen.last_added_ref = project.eigen_referentie

        OTLLogger.logger.debug("Projects global")
        for eigen_ref in HomeDomain.projects.keys():
            OTLLogger.logger.debug(eigen_ref)

        HomeDomain.reload_projects()
        HomeDomain.update_frontend()
        # self.table.reset_ui(self._)

    def reset_ui(self, _):
        self._ = _
        self.new_project_button.setText(self._('new_project_button'))
        self.return_button.setText(self._('return_to_home_screen'))
        self.save_button.setText(self._('save_button'))
        self.import_button.setText(self._("import"))
        self.help_action.setText(self._('help'))
        self.about_action.setText(self._('about'))
        self.report_action.setText(self._('report error'))
        if global_vars.current_project is not None:
            self.reference_title.setText(global_vars.current_project.eigen_referentie)
        else:
            self.reference_title.setText("")

    def set_import_icon(self, button: ButtonWidget):
        button.setIcon(qta.icon("mdi.download",
                                            color=Styling.button_icon_color))
    def update_color_scheme(self):
        self.set_import_icon(self.import_button)


from otlmow_gui.Domain import global_vars


class FDOToolboxNotInstalledError(Exception):
    def __init__(self, language):
        self._ = language

        super().__init__()

    def __str__(self):
        # "FDO toolbox executable could not be found. Most likely because it is not installed (correctly), "
        # "You can rerun the OTL wizard installer with admin privileges to fix this error."
        # "Make sure it is installed in the correct directory so that the following path exists: \n{0}"
        translation = self._("FDOToolbox_not_installed_error_message")
        return translation.format(global_vars.FDO_toolbox_path_str, global_vars.FDO_toolbox_installer_path_str)


from PyQt6.QtWidgets import QWidget, QMessageBox


class NotificationWindow(QMessageBox):

    def __init__(self, message:str, title:str = "notification"):
        self.message = message
        self.title = title
        super().__init__()
        self.init_ui()

    def init_ui(self):
        self.setWindowTitle(self.title)
        self.setText(self.message)



from typing import Optional, Callable

from PyQt6.QtGui import QPixmap
import qtawesome as qta

from PyQt6.QtWidgets import QWidget, QLabel, QHBoxLayout, QVBoxLayout, \
    QLineEdit, QFrame, QPushButton, QSizePolicy
from PyQt6.QtCore import Qt

from otlmow_gui.Domain.ProgramFileStructure import ProgramFileStructure
from otlmow_gui.Domain.project.Project import Project
from otlmow_gui.Domain.step_domain.HomeDomain import HomeDomain
from otlmow_gui.Domain.network.Updater import Updater
from otlmow_gui.GUI.Styling import Styling
from otlmow_gui.GUI.dialog_windows.SuggestUpdateWindow import SuggestUpdateWindow

from otlmow_gui.GUI.header.HeaderBar import HeaderBar
from otlmow_gui.GUI.screens.Home_elements.OverviewTable import OverviewTable, \
    LastAddedProjectHighlightDelegate

from otlmow_gui.GUI.screens.screen_interface.HomeScreenInterface import HomeScreenInterface

IMG_DIR = ProgramFileStructure.get_dynamic_library_path('img')


class HomeScreen(HomeScreenInterface):
    """
    Represents the home screen of the application for managing projects.

    This class provides the user interface for searching and displaying project information,
    including options for exporting data and managing project settings. It integrates various
    UI components such as input fields, buttons, and tables to facilitate user interactions.

    Args:
        language_settings (optional): Language settings for the user interface.

    Attributes:
        project (optional): The current project associated with the home screen.
        main_window (optional): Reference to the main application window.
        main_layout (QVBoxLayout): Layout for organizing UI components vertically.
        search_input_field (QLineEdit): Input field for entering search queries.
        search_message (QLabel): Label for displaying search-related messages.
        table (OverviewTable): Table for displaying project data.
        header (HeaderBar): Header bar for the home screen.
    """


    def __init__(self, language_settings=None):
        super().__init__()
        self._ = language_settings
        self.container_home_screen = QVBoxLayout()

        # self.message_box = MessageBox(self._)
        self.head_wrapper = QFrame()
        self.search_input_field = QLineEdit()
        self.clear_search_bar_button = None
        self.search_message = QLabel()
        self.table = OverviewTable(self._)
        self.table.setItemDelegate(LastAddedProjectHighlightDelegate(self))
        self.header = HeaderBar(language=self._, table=self.table)

        self.last_added_ref = None

        self.create_main_content_ui()
        self.setLayout(self.container_home_screen)
        if Updater.needs_update:
            SuggestUpdateWindow( self._,Updater.local_version,Updater.master_version)

    def create_main_content_ui(self) -> None:
        """
        Creates the main user interface for the home screen.

        This method sets up the header, search bar, data table, and logo, arranging them in a vertical layout.

        The function constructs the main content of the home screen by adding various UI components to the layout. It includes a header bar, a search bar, a data table, and an AWV logo, ensuring proper spacing and alignment for a cohesive appearance.

        :return: None
        """
        # Header
        self.header.construct_header_bar()

        # Search bar
        search_container = QVBoxLayout()
        search_container.addWidget(self.draw_search_bar())
        search_container.setContentsMargins(16, 0, 16, 0)

        # Create the table
        self.table.draw_table()
        table_container = QVBoxLayout()
        table_container.addWidget(self.table)
        table_container.setContentsMargins(16, 0, 16, 0)

        #AWV logo
        AWV_logo_label = QLabel()
        pixmap = QPixmap(f'{IMG_DIR}/Vlaanderen_is_veilig_onderweg_vol-01_200.png')
        AWV_logo_label.setPixmap(pixmap)

        # add header to the vertical layout
        self.container_home_screen.addWidget(self.header)
        self.container_home_screen.addSpacing(39)
        # add searchbar to the vertical layout
        self.container_home_screen.addLayout(search_container)
        self.container_home_screen.addSpacing(43)
        # add table to the vertical layout with margins
        self.container_home_screen.addLayout(table_container)
        # self.container_home_screen.addStretch()
        self.container_home_screen.addSpacing(10)
        self.container_home_screen.addWidget(AWV_logo_label, alignment=Qt.AlignmentFlag.AlignRight)
        self.container_home_screen.setContentsMargins(0, 0, 0, 0)


    def fill_table(self,projects: [Project]) -> None:
        """
        Fills the table with project data based on the provided list of projects.

        This method updates the table display using the current search input to filter the projects.

        The function calls the `fill_table` method of the table component, passing the list of projects and the text from the search input field to ensure that only relevant projects are displayed.

        :param projects: A list of Project instances to populate the table.
        :type projects: list[Project]

        :return: None
        """

        self.table.fill_table(projects, self.search_input_field.text())


    def draw_search_bar(self) -> QWidget:
        """
        Creates and returns a search bar UI component.

        This method constructs a horizontal layout containing an input field for search queries
        and a message label for displaying error messages.

        The function initializes a QWidget as a wrapper for the search bar, applies a CSS class,
        and adds the input field and message label to the layout. The message label is styled to
        display text in red, indicating potential errors or prompts to the user.

        :return: The constructed search bar widget.
        :rtype: QWidget
        """

        search_wrapper = QWidget()
        search_wrapper.setProperty('class', 'search')
        search = QHBoxLayout()
        self.create_input_field()
        self.search_message.setText("")
        self.search_message.setStyleSheet("color: red")

        self.clear_search_bar_button = QPushButton()
        self.set_clear_icon(self.clear_search_bar_button)
        self.clear_search_bar_button.clicked.connect(self.clear_search_listener)
        self.clear_search_bar_button.setProperty('class', 'secondary-button')

        search.addWidget(self.search_input_field)
        search.addWidget(self.clear_search_bar_button)
        search.addWidget(self.search_message)

        search_wrapper.setLayout(search)
        return search_wrapper

    # noinspection PyMethodMayBeStatic
    def set_clear_icon(self, button: QPushButton) -> None:
        button.setIcon(qta.icon('mdi.close', color=Styling.button_icon_color))

    def clear_search_listener(self) -> None:
        self.search_input_field.setText("")
        HomeDomain.update_frontend()

    def create_input_field(self) -> None:
        """
        Initializes the search input field for the search bar.

        This method sets up a placeholder text and connects the return key event to update the frontend when the user submits a search query.

        The function configures the search input field to display a placeholder that prompts the user for input. It also establishes a connection to trigger an update of the frontend when the return key is pressed.

        :return: None
        """

        self.search_input_field.textChanged.connect(lambda: HomeDomain.update_frontend())
        self.search_input_field.setPlaceholderText(self._('search_text'))
        self.search_input_field.setSizePolicy(QSizePolicy.Policy.Expanding,QSizePolicy.Policy.Fixed)


    def remove_table_row(self, row_index: int) -> None:
        """
        Removes a specified row from the table.

        This method updates the table by removing the row at the given index, ensuring that the
        displayed data reflects the change.

        :param row_index: The index of the row to be removed from the table.
        :type row_index: int

        :return: None
        """
        self.table.removeRow(row_index)


    def reset_ui(self, lang_settings: Optional[Callable]=None) -> None:
        """
        Resets the user interface components to their default state.

        This method updates the UI elements, including the table and header, and sets the
        placeholder text for the search input field. It can also accept language settings to
        adjust the UI language if provided.

        :param lang_settings: Optional language settings to apply to the UI.
        :type lang_settings: dict or None

        :return: None
        """
        super().reset_ui(self._)
        if lang_settings is not None:
            self._ = lang_settings

        self.table.reset_ui(self._)
        self.search_input_field.setPlaceholderText(self._('search_text'))
        self.header.reset_ui(self._)
        self.last_added_ref = None
        # self.sort_on_last_edit()

    def sort_on_last_edit(self):
        self.table.activate_initial_sort_on_last_edit()



from typing import Optional

from otlmow_gui.GUI.dialog_windows.file_picker_dialog.SaveFilePickerDialog import SaveFilePickerDialog


class TemplateSaveFilePickerDialog(SaveFilePickerDialog):


    def __init__(self, language_settings, title: Optional[str] = None,
                 name_filter: str = None, initial_file_path: Optional[str] = None,
                 exclude_file_types: list[str] = []):

        exclude_file_types.extend(['GeoJSON',"JSON"])

        super().__init__(language_settings, title, name_filter, initial_file_path,
                         exclude_file_types)

        if not title:
            self.setWindowTitle(self._("Sla template op"))

    def get_filename_suggestion(self, project_name):
        return project_name + "_template"






import importlib.metadata
import json
import os
import shutil
import tomllib
import traceback
from pathlib import Path

import toml

from otlmow_gui.Domain.ProgramFileStructure import ProgramFileStructure
from otlmow_gui.Domain.logger.OTLLogger import OTLLogger
from otlmow_gui.Domain.network.GitHubDownloader import GitHubDownloader
from otlmow_gui.Domain.util.Helpers import Helpers


class Updater:

    pyproject_filename = "pyproject.toml"
    pyproject_config = None
    needs_update = False
    github_downloader = GitHubDownloader('davidvlaminck/OTLMOW-GUI')

    local_version = None
    master_version = None

    @classmethod
    def get_project_version(cls):
        if not cls.pyproject_config:
            pyproject_path = ProgramFileStructure.get_dynamic_library_path(cls.pyproject_filename)
            cls.pyproject_config = toml.load(pyproject_path)

        return cls.pyproject_config['project']['version']

    @classmethod
    def check_for_OTL_wizard_updates(cls):
        # TODO: catch and process ratelimit exceeded error
        try:
            contents = cls.github_downloader.download_file_to_memory("pyproject.toml").decode("utf-8")
            master_pyproject_config = tomllib.loads(contents)

            cls.local_version = cls.get_project_version()
            cls.master_version = master_pyproject_config['project']['version']

            if Helpers.is_version_equal_or_higher(cls.local_version, cls.master_version):
                OTLLogger.logger.info(f"Local version {cls.local_version} is up to date with master version")
            else:
                cls.needs_update = True
                OTLLogger.logger.info(f"Update needed: Local version {cls.local_version} != {cls.master_version} master version" )
        except Exception as e:
            OTLLogger.logger.info(f"Couldn't check if new version is available for update")
            OTLLogger.logger.debug(e)

    @classmethod
    def download_fresh_otlmow_model(cls) -> Path:

        ghdl = GitHubDownloader('davidvlaminck/OTLMOW-Model')
        download_filename = 'full_repo_download.zip'
        temp_download_dest_path = Helpers.create_temp_path(path_to_template_file_and_extension=Path(download_filename))
        temp_base_dir_path = Helpers.get_base_temp_dir_path()

        ghdl.download_full_repo(temp_download_dest_path)

        unzip_dir_path = temp_base_dir_path / 'downloaded_model'
        if unzip_dir_path.exists():
            shutil.rmtree(unzip_dir_path)

        OTLLogger.logger.info(f"unpack downloaded otlmow-model to {unzip_dir_path}",extra={"timing_ref":"unpack_otlmow_model"} )
        shutil.unpack_archive(temp_download_dest_path,
                              unzip_dir_path)
        OTLLogger.logger.info("unpack downloaded otlmow-model", extra={"timing_ref":"unpack_otlmow_model"} )


        return unzip_dir_path

    @classmethod
    def get_otlmow_model_version(cls) -> str:
        ghdl = GitHubDownloader('davidvlaminck/OTLMOW-Model')
        # ghdl.download_file_to_dir(file_path='otlmow_model/version_info.json',
        #                           destination_dir=model_dir_path / 'temp')
        # with open(model_dir_path / 'temp' / 'version_info.json') as json_file:
        #     version_info = json.load(json_file)
        contents = ghdl.download_file_to_memory("otlmow_model/version_info.json").decode("utf-8")
        version_info = json.loads(contents)
        print(f"version info {version_info}")
        if not isinstance(version_info,dict):
            return ""

        try:
            model_version = version_info["current"]['model_version']
        except:
            model_version = ""

        OTLLogger.logger.info(f"online otlmow-model version: {model_version}")
        return model_version

    @classmethod
    def update_oltmow_model(cls, model_dir_path=None):
        OTLLogger.logger.info(
            "Checking for new otlmow-model versions to update to")

        library_name = "otlmow_model"
        otlmow_model_lib_path = ProgramFileStructure.get_dynamic_library_path(library_name)

        if not otlmow_model_lib_path.exists():
            OTLLogger.logger.info("No otlmow-model library directory found to update! are you in testing environment?")
            return
        try:
            local_model_version = cls.get_local_otl_model_library_version()
            online_model_version = cls.get_otlmow_model_version()

            if  (local_model_version and online_model_version and
                    Helpers.is_version_equal_or_higher(local_model_version, online_model_version)):
                OTLLogger.logger.info("otlmow-model is uptodate")
                return

            try:

                OTLLogger.logger.info("Updating otlmow-model", extra={"timing_ref": "update_otlmow_model"})
                unzip_dir_path = cls.download_fresh_otlmow_model()

                files_in_unzip_dir_path_list = os.listdir(unzip_dir_path)
                if files_in_unzip_dir_path_list:

                    new_otlmow_model_dir_path = unzip_dir_path/ files_in_unzip_dir_path_list[0] / 'otlmow_model'

                    # replace old otlmow_model folder in application site_packages with new otlmow_model
                    OTLLogger.logger.info(
                        f"Replacing old otlmow_model folder at: {otlmow_model_lib_path}", extra={"timing_ref": "replace_otl_model"})
                    if otlmow_model_lib_path.exists():
                        shutil.rmtree(otlmow_model_lib_path)

                    shutil.copytree(new_otlmow_model_dir_path,otlmow_model_lib_path)
                    OTLLogger.logger.info(
                        "Replaced old otlmow_model folder", extra={"timing_ref": "replace_otl_model"})
                    OTLLogger.logger.info("Updated otlmow-model",
                                          extra={"timing_ref": "update_otlmow_model"})
                    return

                OTLLogger.logger.info("FAILED Updating otlmow-model", extra={"timing_ref": "update_otlmow_model"})
            except Exception as e:
                OTLLogger.logger.info("FAILED Updating otlmow-model",
                                      extra={"timing_ref": "update_otlmow_model"})
                OTLLogger.logger.error("".join(traceback.format_exception(e)))

        except Exception as e:
            OTLLogger.logger.info("FAILED determining if local otlmow-model library is out of date (probably connection issue)")
            OTLLogger.logger.error("".join(traceback.format_exception(e)))
    @classmethod
    async def async_get_local_otl_model_library_version(cls):

        # first if there is a version file in the otlmow-model library itself
        # this is there if the model was downloaded from github
        library_name = "otlmow_model"
        otlmow_model_lib_path = ProgramFileStructure.get_dynamic_library_path(library_name)
        otlmow_model_lib_version_file_path = otlmow_model_lib_path / "version_info.json"

        if otlmow_model_lib_version_file_path.exists():
            with open(otlmow_model_lib_version_file_path) as version_file:
                otl_version_dict = json.load(version_file)
                otl_model_version = otl_version_dict["current"]["model_version"]
        else:

            # second go off the metadata from the pip install
            metadata = importlib.metadata.metadata("otlmow-model")
            otl_model_version = metadata['Version']
        return otl_model_version

    @classmethod
    def get_local_otl_model_library_version(cls):

        # first if there is a version file in the otlmow-model library itself
        # this is there if the model was downloaded from github
        library_name = "otlmow_model"
        otlmow_model_lib_path = ProgramFileStructure.get_dynamic_library_path(library_name)
        otlmow_model_lib_version_file_path = otlmow_model_lib_path / "version_info.json"

        if otlmow_model_lib_version_file_path.exists():
            with open(otlmow_model_lib_version_file_path) as version_file:
                otl_version_dict = json.load(version_file)
                otl_model_version = otl_version_dict["current"]["model_version"]
                OTLLogger.logger.info(
                    f"the model was downloaded from github {otl_model_version}")
        else:

            # second go off the metadata from the pip install
            metadata = importlib.metadata.metadata("otlmow-model")
            otl_model_version = metadata['Version']
            OTLLogger.logger.info(f"checked second go off the metadata from the pip install {otl_model_version}")
        return otl_model_version


from PyQt6.QtWidgets import QMessageBox, QPushButton

from otlmow_gui.Domain.step_domain.ExportFilteredDataSubDomain import ExportFilteredDataSubDomain


class ChooseFileNameWindow:

    def __init__(self, language_settings, project, new_files):
        self._ = language_settings
        self.project = project
        self.new_files = new_files

    def accept(self, dialog_window):
        file_name = "changed_files.json"
        ExportFilteredDataSubDomain().replace_files_with_diff_report(project=self.project, original_documents=self.new_files,
                                                                     file_name=file_name)
        dialog_window.close()

    def warning_overwrite_screen(self):
        dlg = QMessageBox()
        dlg.setWindowTitle(self._("warning overwriting files"))
        dlg.setText(self._("warning overwriting files text"))
        dlg.setIcon(QMessageBox.Icon.Warning)
        dlg.addButton(QPushButton(self._("yes")), QMessageBox.ButtonRole.YesRole)
        dlg.addButton(QPushButton(self._("no")), QMessageBox.ButtonRole.NoRole)
        dlg.exec()
        reply = dlg.buttonRole(dlg.clickedButton())
        if reply == QMessageBox.ButtonRole.YesRole:
            self.accept(dlg)



from datetime import datetime

from PyQt6.QtWidgets import QTableWidgetItem


class OverviewTableItem(QTableWidgetItem):
    
    date_format: str = "%d-%m-%Y"
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __lt__(self, other):
        try:
            return  datetime.strptime(self.text(), OverviewTableItem.date_format) < datetime.strptime(other.text(), OverviewTableItem.date_format)
        except ValueError:
            return super().__lt__(other)

    def __gt__(self, other):

        try:
            return datetime.strptime(self.text(), OverviewTableItem.date_format) > datetime.strptime(other.text(),
                                                                                     OverviewTableItem.date_format)
        except ValueError:
            return super().__gt__(other)

    def __le__(self, other):
        try:
            return datetime.strptime(self.text(), OverviewTableItem.date_format) < datetime.strptime(other.text(),
                                                                                     OverviewTableItem.date_format)
        except ValueError:
            return super().__le__(other)

    def __ge__(self, other):

        try:
            return datetime.strptime(self.text(), OverviewTableItem.date_format) > datetime.strptime(other.text(),
                                                                                     OverviewTableItem.date_format)
        except ValueError:
            return super().__ge__(other)



from abc import abstractmethod
from copy import deepcopy
from pathlib import Path

from PyQt6.QtWidgets import QVBoxLayout, QPushButton, QFrame, QLabel, QHBoxLayout, QComboBox, \
    QCheckBox

from otlmow_gui.Domain import global_vars
from otlmow_gui.Domain.step_domain.ExportDataDomain import ExportDataDomain
from otlmow_gui.Domain.util.Helpers import Helpers
from otlmow_gui.GUI.dialog_windows.file_picker_dialog.SaveFilePickerDialog import SaveFilePickerDialog
from otlmow_gui.GUI.screens.Screen import Screen
from otlmow_gui.exception_handler.ExceptionHandlers import create_task_reraise_exception


class AbstractExportDataSubScreen(Screen):

    def __init__(self, language_settings=None):
        super().__init__()
        self._ = language_settings

        self.export_btn = QPushButton()
        self.input_file_label = QLabel()
        self.extra_option_csv = QCheckBox()
        self.file_extension_selection = QComboBox()
        self.title = QLabel()
        self.file_type_label = QLabel()
        self.main_window = None
        self.relations_split_optionality = QCheckBox()
        self.supported_export_formats:dict = deepcopy(global_vars.supported_file_formats)
        if "SDF" in self.supported_export_formats:
            self.supported_export_formats.pop("SDF") # not yet supported for export in V0.5.0

        self.export_file_dialog_window = SaveFilePickerDialog(self._)

        self.init_ui()


    @abstractmethod
    def init_ui(self) -> None:
        pass

    def create_filetype_combobox(self) -> QFrame:
        """
        Creates a combo box for selecting the file type for export. This method sets up the layout,
        adds a label and a combo box populated with supported file formats, and connects the c
        ombo box's change event to a method for displaying additional options.

        :return: A QFrame containing the combo box for file type selection.
        :rtype: QFrame
        """

        frame = QFrame()
        frame_layout = QHBoxLayout()

        self.file_type_label.setText(self._('select file type for export') + ":")

        self.file_extension_selection.addItems(list(self.supported_export_formats.keys()))
        self.file_extension_selection.currentTextChanged.connect(self.show_additional_options)

        frame_layout.addWidget(self.file_type_label)
        frame_layout.addWidget(self.file_extension_selection)
        frame_layout.addStretch()
        frame_layout.setContentsMargins(0,0,0,0)
        frame.setLayout(frame_layout)
        return frame

    def create_relations_split_optionality(self):
        checkbox_frame = QFrame()
        checkbox_layout = QVBoxLayout()

        self.relations_split_optionality.setText(
            self._('export relations and assets in different files'))

        checkbox_layout.addWidget(self.relations_split_optionality)
        checkbox_layout.addWidget(self.create_extra_option_csv())
        checkbox_layout.setContentsMargins(11,0,0,0)
        checkbox_frame.setLayout(checkbox_layout)

        return checkbox_frame

    def create_button_box(self):
        """
        Creates a button box containing the export button for the export data screen.
        This method sets up the layout for the button, assigns its properties,
        and connects the button's click event to a method for opening the file picker.

        :return: A QFrame containing the button box with the export button.
        :rtype: QFrame
        """

        button_box = QFrame()
        button_box_layout = QHBoxLayout()

        self.export_btn.setText(self._('export'))
        self.export_btn.setProperty('class', 'primary-button')
        self.export_btn.clicked.connect(lambda: self.open_file_picker())

        button_box_layout.addWidget(self.export_btn)
        button_box_layout.addStretch()
        button_box_layout.setContentsMargins(0, 0, 0, 0)
        button_box.setLayout(button_box_layout)

        return button_box

    def reset_ui(self, _):
        super().reset_ui(_)
        """
        Creates a button box containing the export button for the export data screen. This method
        sets up the layout for the button, assigns its properties, and connects the button's click
        event to a method for opening the file picker.

        :return: A QFrame containing the button box with the export button.
        :rtype: QFrame
        """

        self._ = _
        self.export_btn.setText(self._('export'))
        self.file_type_label.setText(self._('select file type for export') + ":")
        self.relations_split_optionality.setText(self._('export relations and assets in different files'))
        self.extra_option_csv.setText(self._('export assets in different files'))
        self.input_file_label.setText(self._('file_to_upload'))


    def create_extra_option_csv(self):
        self.extra_option_csv.setText(self._('export assets in different files'))
        self.extra_option_csv.setHidden(True)
        return self.extra_option_csv

    def open_file_picker(self):
        """
        Opens a file picker dialog for the user to select a location to save exported files.
        This method configures the dialog based on the currently selected file format and options,
        and triggers the file generation process if a valid file location is chosen.

        :return: None
        """



        chosen_file_format = self.file_extension_selection.currentText()
        if chosen_file_format in self.supported_export_formats:
            document_path_list = self.export_file_dialog_window.summon(
                chosen_file_format=chosen_file_format,
                supported_export_formats=self.supported_export_formats,
                project_name=global_vars.current_project.eigen_referentie)

            if document_path_list and document_path_list[0]:
                csv_option = self.extra_option_csv.isChecked()
                split_relations_and_objects = self.relations_split_optionality.isChecked()
                try:
                    self.process_export(document_path_list, csv_option, split_relations_and_objects)
                except Exception as e:
                    # TODO: proper error message when file fails to be exported
                    raise e

    def process_export(self, document_path_list, csv_option, split_relations_and_objects):
        create_task_reraise_exception(
            ExportDataDomain.generate_files(end_file=document_path_list[0],
                                            separate_per_class_csv_option=csv_option,
                                            separate_relations_option=split_relations_and_objects))

    def show_additional_options(self, text):
        """
        Displays additional options based on the selected file type in the export data screen.
        This method shows or hides the CSV option based on whether the user selects 'CSV' as the
        file type, ensuring that only relevant options are visible.

        :param text: The selected file type that determines which options to display.
        :type text: str

        :return: None
        """

        if text == 'CSV':
            self.extra_option_csv.setHidden(False)
            self.extra_option_csv.setChecked(True)
        else:
            self.extra_option_csv.setChecked(False)
            self.extra_option_csv.setHidden(True)


    def open_folder_of_created_export_files(self, document_path:Path):
        Helpers.open_folder_and_select_document(document_path)